import math

__all__ = [
    "Memory",
    "MemorySubModule",
    "DMNoise",
    "DMBaseLevel",
    "DMSalience",
    "DMSpreading",
    "DMFixed",
    "Partial",
    "BlendingMemory",
    "DMAssociate",
    "DMInhibition",
]

from .buffer import Chunk, Buffer
from .pattern import Pattern
from .model import Model


class Memory(Model):
    def __init__(
        self,
        buffer,
        latency=0.05,
        threshold=0,
        maximum_time=10.0,
        finst_size=4,
        finst_time=3.0,
    ):
        Model.__init__(self)
        self._buffer = buffer
        self.dm = []
        self.error = False
        self.busy = False
        self.adaptors = []
        self.latency = latency
        self.threshold = threshold
        self.maximum_time = maximum_time
        self.partials = []
        self.finst = Finst(self, size=finst_size, time=finst_time)
        self.record_all_chunks = False
        self._request_count = 0

    def clear(self):
        del self.dm[:]

    def add(self, chunk, record=None, **keys):
        if self.error:
            self.error = False
        if isinstance(chunk, Buffer):
            chunk = Chunk(chunk.chunk)
        else:
            bound = None
            if hasattr(self, "sch"):
                bound = getattr(self.sch, "bound", None)
            chunk = Chunk(chunk, bound)
        for c in self.dm:
            if chunk == c:
                for a in self.adaptors:
                    a.merge(c, **keys)
                break
        else:
            for a in self.adaptors:
                a.create(chunk, **keys)
            self.dm.append(chunk)
        chunk.record = record

    def find_matching_chunks(self, pattern, threshold=None):
        bound = getattr(self.sch, "bound", None)
        pattern = Pattern(pattern, bound)
        matches = [x for x in self.dm if pattern.match(x) is not None]
        if threshold is not None:
            matches = [x for x in matches if self.get_activation(x) >= threshold]
        return matches

    def request(self, pattern, partial=None, require_new=False):
        if partial is None and len(self.partials) > 0:
            partial = self.partials[0]
        self.busy = True
        if self.error:
            self.error = False
        self._request_count += 1
        b = getattr(self.sch, "bound", None)
        pattern = Pattern(pattern, b, partial=partial)

        all = self.dm
        if require_new:
            all = [x for x in all if not self.finst.contains(x)]

        matches = [x for x in all if pattern.match(x) is not None]

        # for x in matches: print `x`,self.get_activation(x)

        for a in self.adaptors:
            a.matched(matches)
        if len(matches) == 0:
            self.fail(self._request_count)
        else:
            maximum = None
            for chunk in matches:
                chunk.activation = self.get_activation(chunk)
                if partial is not None:
                    chunk.activation += chunk._partial
                if maximum is None or chunk.activation > maximum:
                    maximum = chunk.activation
            if maximum < self.threshold:
                self.fail(self._request_count)
            else:
                best = [x for x in matches if x.activation == maximum]

                choice = self.random.choice(best)
                self.recall(choice, matches=matches, request_number=self._request_count)

    def fail(self, request_number):
        if self.threshold is None:
            time = self.maximum_time
        else:
            time = self.latency * math.exp(-self.threshold)
            if time > self.maximum_time:
                time = self.maximum_time
        yield time
        if request_number != self._request_count:
            return

        self.error = True
        self._buffer.clear()
        self.busy = False

    def recall(self, chunk, matches, request_number):
        self.finst.add(chunk)
        time = self.latency * math.exp(-chunk.activation)
        if time > self.maximum_time:
            time = self.maximum_time
        yield time
        if request_number != self._request_count:
            return
        self._buffer.set(chunk)
        for a in self.adaptors:
            a.recalled(chunk)
        self.busy = False

    def get_activation(self, chunk):
        if not isinstance(chunk, Chunk):
            try:
                chunk = Chunk(chunk, self.sch.bound)
            except AttributeError:
                chunk = Chunk(chunk, None)
            for c in self.dm:
                if chunk == c:
                    chunk = c
                    break
            else:
                raise Exception("No such chunk found")
        act = 0
        for a in self.adaptors:
            act += a.activation(chunk)
        if self.record_all_chunks or chunk.record is True:
            self.log[str(chunk)] = act
        return act

    def add_adaptor(self, a):
        self.adaptors.append(a)


class Finst:
    def __init__(self, parent, size=4, time=3.0):
        self.parent = parent
        self.size = size
        self.time = time
        self.obj = []

    def contains(self, o):
        return o in self.obj

    def add(self, o):
        if self.size == 0:
            return
        self.obj.append(o)
        if len(self.obj) > self.size:
            self.remove(self.obj[0])
        self.parent.sch.add(self.remove, args=[o], delay=self.time)

    def remove(self, o):
        if o in self.obj:
            self.obj.remove(o)


class MemorySubModule:
    def __init__(self, parent):
        self.parent = parent
        if parent is not None:
            parent.add_adaptor(self)

    def create(self, chunk, **keys):
        pass

    def merge(self, chunk, **keys):
        pass

    def matched(self, chunks):
        pass

    def activation(self, chunk):
        return 0

    def recalled(self, chunk):
        pass

    def now(self):
        if self.parent is None or not self.parent._is_converted():
            return 0
        else:
            return self.parent.now()


class DMNoise(MemorySubModule):
    def __init__(self, memory, noise=0.3, baseNoise=0.0):
        MemorySubModule.__init__(self, memory)
        self.noise = noise
        self.baseNoise = baseNoise

    def create(self, chunk, **keys):
        chunk.baseNoise = self.logisticNoise(self.baseNoise)

    def activation(self, chunk):
        return chunk.baseNoise + self.logisticNoise(self.noise)

    def logisticNoise(self, s):
        try:
            x = self.parent.random.random()
        except AttributeError:
            import random

            x = random.random()
        return s * math.log(1.0 / x - 1.0)


class DMBaseLevel(MemorySubModule):
    def __init__(self, memory, decay=0.5, limit=None):
        MemorySubModule.__init__(self, memory)
        self.decay = decay
        self.limit = limit

    def create(self, chunk, time=0.0, baselevel=None, **keys):
        chunk.creation = self.now()
        chunk.times = [chunk.creation - time]
        chunk.count = 1
        if baselevel is not None and baselevel != "calculate":
            chunk.baselevel = baselevel

    def merge(self, chunk, time=0.0, baselevel=None, **keys):
        chunk.times.append(self.now() - time)
        chunk.count += 1
        if self.limit is not None and len(chunk.times) > self.limit:
            if self.limit == 0:
                del chunk.times[:]
            else:
                chunk.times = chunk.times[-self.limit :]
        if baselevel is not None:
            if baselevel == "calculate":
                del chunk.baselevel
            else:
                chunk.baselevel = baselevel

    def activation(self, chunk):
        if hasattr(chunk, "baselevel"):
            return chunk.baselevel
        if self.decay == None:
            return 0
        d = self.decay
        now = self.now()
        t = [now - time for time in chunk.times]
        t = [max(L, 0.005) for L in t]

        exact = 0.0
        approx = 0.0
        if len(t) > 0:
            exact = sum([math.pow(time, -d) for time in t])
        if self.limit is not None:
            n = chunk.count
            k = len(chunk.times)
            if n > k:
                tn = now - chunk.creation
                if k > 0:
                    tk = now - chunk.times[0]
                else:
                    tk = 0.0
                approx = (
                    (n - k)
                    / (1 - d)
                    * (math.pow(tn, 1 - d) - math.pow(tk, 1 - d))
                    / (tn - tk)
                )

        B = math.log(exact + approx)
        return B


class DMSpacing(MemorySubModule):
    def __init__(self, memory, decayScale=0.0, decayIntercept=0.5):
        MemorySubModule.__init__(self, memory)
        self.decayScale = decayScale
        self.decayIntercept = decayIntercept

    def create(self, chunk, time=0.0, **keys):
        chunk.times = [(self.actr.time - time, self.decayIntercept)]

    def merge(self, chunk, time=0.0, **keys):
        t = self.now() - time
        m = self.activation(chunk)
        d = self.decayScale * math.exp(m) + self.decayIntercept
        chunk.times.append((t, d))

    def activation(self, chunk):
        now = self.now()
        total = 0.0
        for time, d in chunk.times:
            total += math.pow(now - time, -d)
        B = math.log(total)
        return B


class DMInhibition(MemorySubModule):
    def __init__(self, memory, decayScale=1.0, timeScale=5.0):
        MemorySubModule.__init__(self, memory)
        self.decayScale = decayScale
        self.timeScale = timeScale

    def create(self, chunk, time=0.0, **keys):
        chunk.mostRecentTime = time

    def merge(self, chunk, time=0.0, **keys):
        chunk.mostRecentTime = self.now()

    def activation(self, chunk):
        tn = self.now() - chunk.mostRecentTime
        I = -math.log(1 + math.pow(tn, -self.decayScale) / self.timeScale)
        return I


class DMSalience(MemorySubModule):
    def __init__(self, memory):
        MemorySubModule.__init__(self, memory)
        self.histogram = {}
        self.weight = {}
        self.memory = memory

    def weights(self, **weights):
        self.histogram = {}
        self.weight = {}
        for k, v in weights.items():
            if k.startswith("_"):
                k = int(k[1:])
            self.weight[k] = float(v)
            self.histogram[k] = {}

    def context(self, pattern):
        pat = Pattern(pattern)
        chunks = [x for x in self.memory.dm if pat.match(x) is not None]
        for k, hist in self.histogram.items():
            hist.clear()
        if len(chunks) == 0:
            raise Exception('No chunks match salience pattern: "%s"' % pattern)
        dw = 1.0 / len(chunks)
        for c in chunks:
            for k, hist in self.histogram.items():
                val = c.get(k, None)
                if val not in hist:
                    hist[val] = dw
                else:
                    hist[val] += dw

    def activation(self, chunk):
        act = 0.0
        for k, hist in self.histogram.items():
            val = chunk.get(k, None)
            p = hist[val]
            w = self.weight[k]
            act += math.log(1.0 / p, 2) * w
        #    if self.log: self.log.act[`chunk`]=act
        return act


class DMSpreading(MemorySubModule):
    def __init__(self, memory, *buffers):
        MemorySubModule.__init__(self, memory)
        self.strength = 1
        self.buffers = buffers
        self.weight = {}
        for b in buffers:
            self.weight[b] = 1
        self.slots = {}

    def create(self, chunk, **keys):
        for slot in chunk.values():
            if slot in self.slots:
                self.slots[slot].append(chunk)
            else:
                self.slots[slot] = [chunk]

    def activation(self, chunk):
        values = chunk.values()

        total = 0.0
        for b in self.buffers:
            ch = b.chunk
            if ch is not None:
                w = self.weight[b]
                for key, slot in ch.items():
                    if slot in values:
                        s = self.strength - math.log(len(self.slots[slot]) + 1)
                        total += w * s
        return total


class DMFixed(MemorySubModule):
    def __init__(self, memory, default=0):
        MemorySubModule.__init__(self, memory)
        self.default = default

    def create(self, chunk, fixed=None, **keys):
        if fixed is None:
            fixed = self.default
        chunk.fixed = fixed

    def merge(self, chunk, fixed=0, **keys):
        chunk.fixed += fixed

    def activation(self, chunk):
        return chunk.fixed


class Associated:
    def __init__(self, a, b):
        self.a = a
        self.b = b


class DMAssociate(MemorySubModule):
    def __init__(self, memory, buffer, weight=1, decay=0.5, limit=None):
        MemorySubModule.__init__(self, memory)
        self._buffer = buffer
        self._mem = {}
        self._bl = DMBaseLevel(None, decay=decay, limit=limit)
        self.weight = weight

    def set_association(self, pre, post, baselevel):
        c = self._mem.get((pre, post), None)
        if c == None:
            c = Associated(pre, post)
            self._bl.create(c)
            self._mem[(pre, post)] = c
        c.baselevel = baselevel

    def recalled(self, chunk):
        prechunk = self._buffer.chunk
        if prechunk is None:
            return
        for pk, pv in prechunk.items():
            for k, v in chunk.items():
                c = self._mem.get((pv, v), None)
                if c == None:
                    c = Associated(pv, v)
                    self._bl.create(c)
                    self._mem[(pv, v)] = c
                else:
                    self._bl.merge(c)

    def activation(self, chunk):
        act = 0
        prechunk = self._buffer.chunk
        if prechunk is None:
            return 0
        for pk, pv in prechunk.items():
            for k, v in chunk.items():
                c = self._mem.get((pv, v), None)
                if c is not None:
                    act += self._bl.activation(c)
        return act * self.weight


class Partial:
    def __init__(self, memory, strength=1.0, limit=-1.0):
        self._memory = memory
        self.limit = limit
        self.strength = strength
        self.sims = {}
        memory.partials.append(self)

    def similarity(self, a, b, value):
        self.sims[a, b] = value
        self.sims[b, a] = value

    def request(self, pattern):
        self._memory.request(pattern, partial=self)

    def match(self, key, a, b):
        m = self.sims.get((a, b), self.limit)
        p = self.strength
        return p * m


class BlendingMemory(Memory):
    def recall(self, chunk, matches, request_number):
        keys = list(sorted(chunk.keys()))
        if not hasattr(chunk, "_blend_keys"):
            bk = []
            for k in keys:
                try:
                    v = float(chunk[k])
                    bk.append(k)
                except:
                    pass
            chunk._blend_keys = tuple(bk)
        bk = chunk._blend_keys
        if len(bk) > 0:
            a = chunk.activation
            chunk = Chunk(chunk)
            chunk.activation = a

            for b in bk:
                chunk[b] = 0
            total_activation = 0
            for m in matches:
                m.exp_activation = math.exp(m.activation)
                k = list(sorted(m.keys()))
                if k == keys:
                    blend = {}
                    try:
                        for b in bk:
                            blend[b] = float(m[b])
                    except ValueError:
                        continue
                    for b in bk:
                        chunk[b] += blend[b] * m.exp_activation
                    total_activation += m.exp_activation
            for b in bk:
                chunk[b] /= total_activation
        for m in Memory.recall(self, chunk, matches, request_number):
            yield m



"""The reference implementation of the MDF execution engine; allows for executing :class:`~modeci.mdf.Graph`
objects in Python.

This module implements a set of classes for executing loaded MDF models in Python.
The implementation is organized such that each class present in :mod:`~modeci_mdf.mdf`
has a corresponding :code:`Evaluable` version of the class. Each of these classes implements
the execution of these components and tracks their state during execution. The organization of the entire execution of
the model is implemented at the top-level :func:`~modeci_mdf.execution_engine.EvaluableGraph.evaluate` method
of the :class:`EvaluableGraph` class. The external library `graph-scheduler
<https://pypi.org/project/graph-scheduler/>`_ is used to implement the scheduling of nodes under declarative
conditional constraints.

"""
import copy
import functools
import inspect
import itertools
import os
import re
import sys
import math

import attr
import sympy
import numpy as np

import graph_scheduler
import onnxruntime

from modeci_mdf.functions.standard import mdf_functions, create_python_expression
from modeci_mdf.utils import is_number

from modelspec.utils import evaluate as evaluate_params_modelspec
from modelspec.utils import _params_info, _val_info
from modelspec.utils import FORMAT_NUMPY

from collections import OrderedDict
from typing import Union, List, Dict, Optional, Any, Tuple
from modeci_mdf.mdf import (
    Function,
    Graph,
    Condition,
    Edge,
    OutputPort,
    InputPort,
    Node,
    Parameter,
)

import modeci_mdf.functions.onnx as onnx_ops
import modeci_mdf.functions.actr as actr_funcs


FORMAT_DEFAULT = FORMAT_NUMPY

KNOWN_PARAMETERS = ["constant"]


time_scale_str_regex = r"(TimeScale)?\.(.*)"


def evaluate_expr(
    expr: Union[str, List[str], np.ndarray, "tf.tensor"] = None,
    func_params: Dict[str, Any] = None,
    array_format: str = FORMAT_DEFAULT,
    verbose: Optional[bool] = False,
) -> np.ndarray:

    """Evaluates an expression given in string format and a :code:`dict` of parameters.

    Args:
        expr: Expression or list of expressions to be evaluated
        func_params: A dict of parameters (e.g. :code:`{'weight': 2}`)
        array_format: It can be a n-dimensional array or a tensor
        verbose: If set to True provides in-depth information else verbose message is not displayed

    Returns:
        n-dimensional array

    """

    e = evaluate_params_modelspec(
        expr, func_params, array_format=array_format, verbose=verbose
    )
    if type(e) == str and e not in KNOWN_PARAMETERS:
        raise Exception(
            "Error! Could not evaluate expression [%s] with params %s, returned [%s] which is a %s"
            % (expr, _params_info(func_params, multiline=True), e, type(e))
        )
    return e


def evaluate_onnx_expr(
    expr: str,
    base_parameters: Dict[str, Any],
    evaluated_parameters: Dict,
    verbose: bool = False,
) -> Any:
    """Evaluates a simple expression in string format representing an
    ONNX function call

    Args:
        expr (str): Expression to be evaluated
        base_parameters (Dict[str, Any]): A dict of parameters that may contain variables
        evaluated_parameters (Dict): A dict mapping variables used in **base_parameters** to actual values
        verbose (bool, optional): If set to True provides in-depth information else verbose message is not displayed. Defaults to False.

    Returns:
        Any: the return value of **expr**
    """
    # Get the ONNX function
    onnx_name = expr.split("(")[0].split(".")[-1]
    onnx_function = getattr(onnx_ops, onnx_name)
    onnx_schema = onnx_ops.get_onnx_schema(onnx_name)
    onnx_arguments = set(
        list(onnx_schema.attributes.keys()) + [i.name for i in onnx_schema.inputs]
    )
    # used to attempt to match inputs to expected onnx input types
    onnx_typecast_mappings = {
        onnx_schema.AttrType.INT: int,
        onnx_schema.AttrType.FLOAT: float,
        onnx_schema.AttrType.STRING: str,
        onnx_schema.AttrType.INTS: functools.partial(np.array, dtype=int),
        onnx_schema.AttrType.FLOATS: functools.partial(np.array, dtype=float),
        onnx_schema.AttrType.STRINGS: functools.partial(np.array, dtype=str),
        # TODO: add tensor and graph types?
    }

    try:
        has_variadic = (
            onnx_schema.inputs[0].option == onnx_schema.FormalParameterOption.Variadic
        )
    except IndexError:
        has_variadic = False

    # ONNX functions expect input args or kwargs first, followed by parameters (called attributes in ONNX) as
    # kwargs. Lets construct this.
    kwargs_for_onnx = {}
    for kw, arg_expr in base_parameters.items():

        # If this arg is a list of args, we are dealing with a variadic argument. Expand these
        if type(arg_expr) == str and arg_expr[0] == "[" and arg_expr[-1] == "]":
            # Use the Python interpreter to parse this into a List[str]
            arg_expr_list = parse_str_as_list(arg_expr[1:-1])
            kwargs_for_onnx.update({a: evaluated_parameters[a] for a in arg_expr_list})
        else:
            kwargs_for_onnx[kw] = evaluated_parameters[kw]

    kwargs_for_onnx = {
        k: v
        for k, v in kwargs_for_onnx.items()
        if (
            (k in onnx_arguments or has_variadic)
            and "onnx::" not in k  # filter Evaluable__ class names
        )
    }

    # attempt to cast attributes to what onnx_function expects
    for k, v in kwargs_for_onnx.items():
        try:
            onnx_attr = onnx_schema.attributes[k]
        except KeyError:
            continue

        try:
            cast_type = onnx_typecast_mappings[onnx_attr.type]
        except KeyError:
            continue

        try:
            kwargs_for_onnx[k] = cast_type(v)
        except (TypeError, ValueError):
            pass

    if verbose:
        print(f"Evaluating ONNX function {onnx_name} with {kwargs_for_onnx}")

    try:
        result = onnx_function(**kwargs_for_onnx)
    except (
        onnxruntime.capi.onnxruntime_pybind11_state.NotImplemented,
        onnxruntime.capi.onnxruntime_pybind11_state.Fail,
    ) as e:
        err = str(e)
        if (
            "bound to different types (tensor(double) and tensor(float)" not in err
            and "Could not find an implementation for the node" not in err
        ):
            raise

        # assume this is related to lack of support for float64/double
        # for Cos, Relu (and likely others) on onnx CPUExecutionProvider
        result = onnx_function(
            **{
                k: v.astype(np.float32)
                if hasattr(v, "dtype") and v.dtype == np.float64
                else v
                for k, v in kwargs_for_onnx.items()
            }
        )

    try:
        if result.dtype == np.float32:
            result = result.astype(np.float64)
    except AttributeError:
        pass

    return result


def parse_str_as_list(s: str) -> list:
    """Produces a list from its string representation. Runs `eval` on
    non-list elements within

    Args:
        s (str): a string representing a list

    Returns:
        list: a list containing elements from **s**
    """

    def try_eval_str(t):
        try:
            return eval(t)
        except (NameError, SyntaxError):
            return t

    def _parse_str_as_list(t):
        res = []
        i = 0
        j = 0
        depth = 0

        t = t.replace(" ", "")

        while j < len(t):
            if t[j] == "[":
                depth += 1

            if t[j] == "]":
                depth -= 1

                if depth == 0:
                    res.append(_parse_str_as_list(t[i + 1 : j]))
                    # also passes this check if at end of string
                    if t[j : j + 1] not in ",]":
                        raise ValueError(f"Malformed input at index {j} of {t} {t[j]}")
                    i = j + 1

            if t[j] == ",":
                if depth == 0:
                    if j > i:
                        res.append(try_eval_str(t[i:j]))
                    i = j + 1

            j = j + 1

        if depth > 0:
            raise ValueError(f"Unmatched opening bracket in {t}")
        elif depth < 0:
            raise ValueError(f"Unmatched closing bracket in {t}")

        if j > i:
            res.append(try_eval_str(t[i:j]))

        return res

    res = _parse_str_as_list(s)

    # avoid adding extra unnecessary list
    if len(res) == 1:
        return res[0]
    else:
        return res


class EvaluableFunction:
    """Evaluates a :class:`~modeci_mdf.mdf.Function` value during MDF graph execution.

    Args:
        function: :func:`~modeci_mdf.mdf.Function` to be evaluated e.g. mdf standard function
        verbose: If set to True Provides in-depth information else verbose message is not displayed
    """

    def __init__(self, function: Function = False, verbose: Optional[bool] = False):
        self.verbose = verbose
        self.function = function

    def evaluate(
        self,
        parameters: Dict[str, Any] = None,
        array_format: str = FORMAT_DEFAULT,
    ) -> Dict[str, Any]:

        r"""Performs evaluation on the basis of given parameters and array_format

        Args:
            parameters: A dictionary of function parameters,e.g.logistic, parameters={'gain': 2,"bias": 3,"offset": 1}
            array_format: It can be a n-dimensional array or a tensor

        Returns:
             value of function after evaluation in Dictionary

        """

        expr = None

        # print("functions value and function>>>", self.function.value, self.function.function)

        # func_val  = self.function.value

        if self.function.function:
            for f in mdf_functions:
                if f == self.function.function:
                    expr = create_python_expression(
                        mdf_functions[f]["expression_string"]
                    )
                    break

        if expr is None:
            expr = self.function.value
        #     #raise "Unknown function: {}. Known functions: {}".format(
        #     #    self.function.function,
        #     #    mdf_functions.keys,
        #     #)

        func_params = {}
        func_params.update(parameters)
        if self.verbose:
            print(
                "    Evaluating %s with %s, i.e. [%s]"
                % (self.function, _params_info(func_params), expr)
            )

        if self.function.args is not None:
            for arg in self.function.args:
                func_params[arg] = evaluate_expr(
                    self.function.args[arg],
                    func_params,
                    verbose=False,
                    array_format=array_format,
                )
                if self.verbose:
                    print(
                        "      Arg: {} became: {}".format(
                            arg, _val_info(func_params[arg])
                        )
                    )

        # If this is an ONNX operation, evaluate it without modelspec.

        if "onnx_ops." in expr:
            if self.verbose:
                print(f"{self.function.id} is evaluating ONNX function {expr}")
            self.curr_value = evaluate_onnx_expr(
                expr,
                # parameters get overridden by self.function.args
                {**parameters, **self.function.args},
                func_params,
                self.verbose,
            )
        elif "actr." in expr:
            actr_function = getattr(actr_funcs, expr.split("(")[0].split(".")[-1])
            self.curr_value = actr_function(
                *[func_params[arg] for arg in self.function.args]
            )
        else:
            self.curr_value = evaluate_expr(
                expr, func_params, verbose=self.verbose, array_format=array_format
            )

        if self.verbose:
            print(
                "    Evaluated %s with %s =\t%s"
                % (self.function, _params_info(func_params), _val_info(self.curr_value))
            )
        return self.curr_value


class EvaluableParameter:
    """
    Evaluates the current value of a :class:`~modeci_mdf.mdf.Parameter` during MDF graph execution.

    Args:
        parameter: The parameter to evaluate during execution.
        verbose: Whether to print output of parameter calculations.
    """

    DEFAULT_INIT_VALUE = 0  # Temporary!

    def __init__(self, parameter: Parameter, verbose: bool = False):
        self.verbose = verbose
        self.parameter = parameter

        """
        if self.parameter.default_initial_value is not None:
            if is_number(self.parameter.default_initial_value):

                self.curr_value = float(self.parameter.default_initial_value)

            else:
                try:
                    self.curr_value = evaluate_expr(
                        self.parameter.default_initial_value,
                        None,
                    )
                # evaluate_expr raises Exception
                except Exception:
                    self.curr_value = self.parameter.default_initial_value
        else:
            self.curr_value = None"""
        self.curr_value = None

    def get_current_value(
        self, parameters: Dict[str, Any], array_format: str = FORMAT_DEFAULT
    ) -> Any:
        """
        Get the current value of the parameter; evaluates the expression if needed.

        Args:
            parameters: a dictionary of parameters and their values that may or may not be needed to evaluate this
                parameter.
            array_format: The array format to use (either :code:`'numpy'` or :code:`tensorflow'`).

        Returns:
            The evaluated value of the parameter.
        """
        # FIXME: Shouldn't this just call self.evaluate, seems like there is redundant code here?
        if self.curr_value is None:
            if (
                self.parameter.value is not None
                or self.parameter.default_initial_value is not None
            ):
                if self.parameter.is_stateful():
                    if self.verbose:
                        print(f"    Initial eval of <{self.parameter.summary()}>  ")

                    if self.parameter.default_initial_value is not None:
                        return evaluate_expr(
                            self.parameter.default_initial_value,
                            parameters,
                            verbose=False,
                            array_format=array_format,
                        )
                    else:
                        return self.DEFAULT_INIT_VALUE
                else:
                    ips = {}
                    ips.update(parameters)
                    ips[self.parameter.id] = self.DEFAULT_INIT_VALUE
                    self.curr_value = evaluate_expr(
                        self.parameter.value,
                        ips,
                        verbose=False,
                        array_format=array_format,
                    )
                    if self.verbose:
                        print(
                            "    Initial eval of <{}> = {} ".format(
                                self.parameter, self.curr_value
                            )
                        )

        return self.curr_value

    def evaluate(
        self,
        parameters: Dict[str, Any],
        time_increment: Optional[float] = None,
        array_format: str = FORMAT_DEFAULT,
    ) -> Any:
        """
        Evaluate the parameter and store the result in the :code:`curr_value` attribute.

        Args:
            parameters: a dictionary of parameters and their values that may or may not be needed to evaluate this
                parameter.
            time_increment: a floating point value specifying the timestep size, only used for :code:`time_derivative`
                parameters
            array_format: The array format to use (either :code:`'numpy'` or :code:`tensorflow'`).

        Returns:
            The current value of the parameter.
        """
        if self.verbose:
            print(
                "    Evaluating {} with {} ".format(
                    self.parameter.summary(), _params_info(parameters)
                )
            )

        continue_eval = True
        if len(self.parameter.conditions) > 0:
            for condition in self.parameter.conditions:
                test = evaluate_expr(
                    condition.test,
                    parameters,
                    verbose=False,
                    array_format=array_format,
                )
                if test == True:
                    self.curr_value = evaluate_expr(
                        condition.value,
                        parameters,
                        verbose=False,
                        array_format=array_format,
                    )
                    continue_eval = False
                    break

                if self.verbose:
                    print(
                        " --- %s: %s = %s (continuing: %s)"
                        % (condition.id, condition.test, test, continue_eval)
                    )

        if continue_eval:
            if self.parameter.value is not None:

                self.curr_value = evaluate_expr(
                    self.parameter.value,
                    parameters,
                    verbose=False,
                    array_format=array_format,
                )
            elif self.parameter.function:
                expr = None
                for f in mdf_functions:
                    if self.parameter.function == f:
                        expr = create_python_expression(
                            mdf_functions[f]["expression_string"]
                        )
                if not expr:

                    expr = self.parameter.function
                    # raise "Unknown function: {}. Known functions: {}".format(
                    #    self.parameter.function,
                    #    mdf_functions.keys,
                    # )

                func_params = {}
                func_params.update(parameters)
                if self.verbose:
                    print(
                        "    Evaluating %s with %s, i.e. [%s]"
                        % (self.parameter, _params_info(func_params), expr)
                    )
                for arg in self.parameter.args:
                    func_params[arg] = evaluate_expr(
                        self.parameter.args[arg],
                        func_params,
                        verbose=False,
                        array_format=array_format,
                    )
                    if self.verbose:
                        print(
                            "      Arg: {} became: {}".format(
                                arg, _val_info(func_params[arg])
                            )
                        )

                # If this is an ONNX operation, evaluate it without modelspec.
                if "onnx_ops." in expr:
                    if self.verbose:
                        print(f"{self.parameter.id} is evaluating ONNX function {expr}")
                    self.curr_value = evaluate_onnx_expr(
                        expr,
                        # parameters get overridden by self.parameter.args
                        {**parameters, **self.parameter.args},
                        func_params,
                        self.verbose,
                    )
                elif "actr." in expr:
                    actr_function = getattr(
                        actr_funcs, expr.split("(")[0].split(".")[-1]
                    )
                    self.curr_value = actr_function(
                        *[func_params[arg] for arg in self.parameter.args]
                    )
                else:
                    self.curr_value = evaluate_expr(
                        expr,
                        func_params,
                        verbose=self.verbose,
                        array_format=array_format,
                    )
            else:
                if time_increment == None:

                    self.curr_value = evaluate_expr(
                        self.parameter.default_initial_value,
                        parameters,
                        verbose=False,
                        array_format=array_format,
                    )

                else:
                    td = evaluate_expr(
                        self.parameter.time_derivative,
                        parameters,
                        verbose=False,
                        array_format=array_format,
                    )

                    self.curr_value += td * time_increment

        if self.verbose:
            print(
                "    Evaluated %s with %s \n       =\t%s"
                % (
                    self.parameter.summary(),
                    _params_info(parameters),
                    _val_info(self.curr_value),
                )
            )

        return self.curr_value


class EvaluableOutput:
    r"""Evaluates the current value of an :class:`~modeci_mdf.mdf.OutputPort` during MDF graph execution.

    Args:
        output_port: Attribute of a Node which exports information to the dependent Node object
        verbose: If set to True Provides in-depth information else verbose message is not displayed
    """

    def __init__(self, output_port: OutputPort, verbose: Optional[bool] = False):
        self.verbose = verbose
        self.output_port = output_port
        self.curr_value = None

    def evaluate(
        self,
        parameters: Dict[str, Any] = None,
        array_format: str = FORMAT_DEFAULT,
    ) -> Union[int, np.ndarray]:

        """Evaluate the value at the output port on the basis of parameters and array_format

        Args:
            parameters: Dictionary of global parameters of the Output Port
            array_format: It is a n-dimensional array

        Returns:
            value at output port
        """
        if self.verbose:
            print(
                "    Evaluating %s with %s "
                % (self.output_port, _params_info(parameters))
            )
        self.curr_value = evaluate_expr(
            self.output_port.value, parameters, verbose=False, array_format=array_format
        )

        if self.verbose:
            print(
                "    Evaluated %s with %s \n       =\t%s"
                % (
                    self.output_port,
                    _params_info(parameters),
                    _val_info(self.curr_value),
                )
            )
        return self.curr_value


class EvaluableInput:
    """Evaluates input value at the :class:`~modeci_mdf.mdf.InputPort` of the node during MDF graph execution.

    Args:
        input_port: The :class:`~modeci_mdf.mdf.InputPort` is an attribute of a Node which imports information to the
            :class:`~modeci_mdf.mdf.Node`
        verbose: If set to True Provides in-depth information else verbose message is not displayed
    """

    def __init__(self, input_port: InputPort, verbose: Optional[bool] = False):
        self.verbose = verbose
        self.input_port = input_port
        self.curr_value = 0

    def set_input_value(self, value: Union[str, int, np.ndarray]):
        """Set a new value at input port

        Args:
            value: Value to be set at Input Port
        """
        if self.verbose:
            print(f"    Input value in {self.input_port.id} set to {_val_info(value)}")
        self.curr_value = value

    def evaluate(
        self, parameters: Dict[str, Any] = None, array_format: str = FORMAT_DEFAULT
    ) -> Union[int, np.ndarray]:

        """Evaluates value at Input port based on parameters and array_format

        Args:
            parameters: Dictionary of  parameters
            array_format: It is a n-dimensional array

        Returns:
            value at Input port
        """
        if self.verbose:
            print(
                "    Evaluated %s with params %s =\t%s"
                % (
                    self.input_port,
                    _params_info(parameters),
                    _val_info(self.curr_value),
                )
            )
        return self.curr_value


class EvaluableNode:
    r"""Evaluates a :class:`~modeci_mdf.mdf.Node` during MDF graph execution.

    Args:
        node: A self contained unit of evaluation receiving input from other :class:`~modeci_mdf.mdf.Node`\(s) on
            :class:`~modeci_mdf.mdf.InputPort`\(s).
        verbose: If set to True Provides in-depth information else verbose message is not displayed
    """

    def __init__(self, node: Node, verbose: Optional[bool] = False):
        self.verbose = verbose
        self.node = node
        self.evaluable_inputs = {}
        self.evaluable_parameters = OrderedDict()
        self.evaluable_functions = OrderedDict()

        self.evaluable_outputs = {}

        all_known_vars = []

        all_known_vars += KNOWN_PARAMETERS

        for ip in node.input_ports:
            rip = EvaluableInput(ip, self.verbose)
            self.evaluable_inputs[ip.id] = rip
            all_known_vars.append(ip.id)
            # params_init[ip] = ip.curr_value

        for p in node.parameters:
            all_known_vars.append(p.id)

        """
        for p in node.parameters:
            ep = EvaluableParameter(p, self.verbose)
            self.evaluable_parameters[p.id] = ep
            all_known_vars.append(p.id)
            # params_init[s] = s.curr_value"""

        all_funcs = [f for f in node.functions]

        # Order the functions into the correct sequence
        while len(all_funcs) > 0:
            f = all_funcs.pop(0)  # pop first off list
            if verbose:
                print(
                    "    Checking whether function: %s with args %s is sufficiently determined by known vars %s"
                    % (f.id, f.args, all_known_vars)
                )
            all_req_vars = []
            if f.args:
                for arg in f.args:
                    arg_expr = f.args[arg]

                    # some non-expression/str types will crash in sympy.simplify
                    if not isinstance(arg_expr, (sympy.Expr, str)):
                        continue

                    # If we are dealing with a list of symbols, each must treated separately
                    if (
                        type(arg_expr) == str
                        and arg_expr[0] == "["
                        and arg_expr[-1] == "]"
                    ):
                        # Use the Python interpreter to parse this into a List[str]
                        arg_expr_list = parse_str_as_list(arg_expr)
                    else:
                        arg_expr_list = [arg_expr]

                    for e in arg_expr_list:
                        func_expr = sympy.simplify(e)
                        all_req_vars.extend([str(s) for s in func_expr.free_symbols])

            all_present = [v in all_known_vars for v in all_req_vars]

            if verbose:
                print(
                    "    Are all of %s in %s? %s"
                    % (all_req_vars, all_known_vars, all_present)
                )
            if all(all_present):
                rf = EvaluableFunction(f, self.verbose)
                self.evaluable_functions[f.id] = rf
                all_known_vars.append(f.id)
            #     params_init[f] = self.evaluable_functions[f.id].evaluate(
            #     params_init, array_format=FORMAT_DEFAULT
            # )
            else:
                if len(all_funcs) == 0:
                    raise Exception(
                        "Error! Could not evaluate function: %s with args %s using known vars %s"
                        % (f.id, f.args, all_known_vars)
                    )
                else:
                    all_funcs.append(f)

        all_params_to_check = [p for p in node.parameters]

        if self.verbose:
            print("all_params_to_check: %s" % all_params_to_check)

        # Order the parameters into the correct sequence
        while len(all_params_to_check) > 0:
            p = all_params_to_check.pop(0)  # pop first off list

            if verbose:
                print(
                    "    Checking whether parameter: %s with args: %s, value: %s (%s) is sufficiently determined by known vars %s"
                    % (p.id, p.args, p.value, type(p.value), all_known_vars)
                )
            all_req_vars = []

            if p.value is not None and type(p.value) == str:
                if p.value[0] == "[" and p.value[-1] == "]":
                    # Use the Python interpreter to parse this into a List[str]
                    arg_expr_list = parse_str_as_list(p.value)
                else:
                    arg_expr_list = [p.value]

                for e in arg_expr_list:
                    param_expr = sympy.simplify(e)
                    all_req_vars.extend([str(s) for s in param_expr.free_symbols])

            if p.args is not None:
                for arg in p.args:
                    arg_expr = p.args[arg]

                    # If we are dealing with a list of symbols, each must treated separately
                    if (
                        type(arg_expr) == str
                        and arg_expr[0] == "["
                        and arg_expr[-1] == "]"
                    ):
                        # Use the Python interpreter to parse this into a List[str]
                        arg_expr_list = parse_str_as_list(arg_expr)
                    else:
                        arg_expr_list = [arg_expr]

                    for e in arg_expr_list:
                        param_expr = sympy.simplify(e)
                        all_req_vars.extend([str(s) for s in param_expr.free_symbols])

            all_known_vars_plus_this = all_known_vars + [p.id]
            all_present = [v in all_known_vars_plus_this for v in all_req_vars]

            if verbose:
                print(
                    "    Are all of %s in %s? %s, i.e. %s"
                    % (
                        all_req_vars,
                        all_known_vars_plus_this,
                        all_present,
                        all(all_present),
                    )
                )
            if all(all_present):
                ep = EvaluableParameter(p, self.verbose)
                self.evaluable_parameters[p.id] = ep
                all_known_vars.append(p.id)

            else:
                if len(all_params_to_check) == 0:
                    raise Exception(
                        "Error! Could not evaluate parameter: %s with args %s using known vars %s"
                        % (p.id, p.args, all_known_vars_plus_this)
                    )
                else:
                    all_params_to_check.append(p)  # Add back to end of list...

        for op in node.output_ports:
            rop = EvaluableOutput(op, self.verbose)
            self.evaluable_outputs[op.id] = rop

    def evaluate(
        self,
        time_increment: Union[int, float] = None,
        array_format: str = FORMAT_DEFAULT,
    ):
        """
        Evaluate the Node for one time-step

        Args:
            time_increment: The time-increment to use for this evaluation.
            array_format: The format to use for arrays.

        """

        if self.verbose:
            print(
                "\n  ---------------\n  Evaluating Node: %s with %s"
                % (self.node.id, [p.id for p in self.node.parameters])
            )
        curr_params = {}

        for eip in self.evaluable_inputs:
            i = self.evaluable_inputs[eip].evaluate(
                curr_params, array_format=array_format
            )
            curr_params[eip] = i

        # First set params to previous parameter values for use in funcs and states...
        for ep in self.evaluable_parameters:

            curr_params[ep] = self.evaluable_parameters[ep].get_current_value(
                curr_params, array_format=array_format
            )

        for ef in self.evaluable_functions:
            curr_params[ef] = self.evaluable_functions[ef].evaluate(
                curr_params, array_format=array_format
            )

        # Now evaluate and set params to new parameter values for use in output...
        for ep in self.evaluable_parameters:
            curr_params[ep] = self.evaluable_parameters[ep].evaluate(
                curr_params, time_increment=time_increment, array_format=array_format
            )

        for eop in self.evaluable_outputs:
            self.evaluable_outputs[eop].evaluate(curr_params, array_format=array_format)

    def get_output(self, id: str = None) -> Union[int, np.ndarray, Tuple]:
        """Get value at output port for given output port's id

        Args:
            id: Unique identifier of the output port. If None, return a tuple for all output ports.

        Returns:
            value at the output port. If id is None, return all outputs as a tuple. If there is only
            one output, return just its value.

        """
        if id is not None:
            for rop in self.evaluable_outputs:
                if rop == id:
                    return self.evaluable_outputs[rop].curr_value
        else:
            outputs = tuple(
                self.evaluable_outputs[rop].curr_value for rop in self.evaluable_outputs
            )
            if len(outputs) == 1:
                return outputs[0]
            else:
                return outputs


class EvaluableGraph:
    r"""
    Evaluates a :class:`~modeci_mdf.mdf.Graph` with the MDF execution engine. This is the top-level interface to the execution engine.

    Args:
        graph: A directed graph consisting of :class:`~modeci_mdf.mdf.Node`\(s) connected via :class:`~modeci_mdf.mdf.Edge`\(s)
        verbose: If set to True Provides in-depth information else verbose message is not displayed

    """

    def __init__(self, graph: Graph, verbose: Optional[bool] = False):
        self.verbose = verbose
        print("\nInit graph: %s" % graph.id)
        self.graph = graph
        self.enodes = {}
        self.root_nodes = []
        self.output_nodes = []

        # Get the root (input nodes) of the graph. We will assume all are root nodes
        # and then remove those that have edges to them.
        for node in graph.nodes:
            if self.verbose:
                print("\n  Init node: %s" % node.id)
            en = EvaluableNode(node, self.verbose)
            self.enodes[node.id] = en
            self.root_nodes.append(node.id)

        for edge in graph.edges:
            if (
                edge.receiver in self.root_nodes
            ):  # It could have been already removed...
                self.root_nodes.remove(edge.receiver)

        self.ordered_edges = []
        evaluated_nodes = []
        for rn in self.root_nodes:
            evaluated_nodes.append(rn)

        edges_to_eval = [edge for edge in self.graph.edges]

        while len(edges_to_eval) > 0:
            edge = edges_to_eval.pop(0)
            if edge.sender not in evaluated_nodes:
                edges_to_eval.append(edge)  # Add back to end of list...
            else:
                self.ordered_edges.append(edge)
                evaluated_nodes.append(edge.receiver)

        if self.graph.conditions is not None:
            if self.graph.conditions.node_specific is None:
                conditions = {}
            else:
                conditions = {
                    self.graph.get_node(node): self.parse_condition(cond)
                    for node, cond in self.graph.conditions.node_specific.items()
                }

            termination_conds = {}
            if self.graph.conditions.termination is not None:
                for scale, cond in self.graph.conditions.termination.items():
                    cond = self.parse_condition(cond)

                    # check for TimeScale in form of enum or equivalent unambiguous strings
                    try:
                        scale = re.match(time_scale_str_regex, scale).groups()[1]
                    except (AttributeError, IndexError, TypeError):
                        pass

                    try:
                        termination_conds[graph_scheduler.TimeScale[scale]] = cond
                    except KeyError:
                        termination_conds[scale] = cond
        else:
            conditions = {}
            termination_conds = {}
        self.scheduler = graph_scheduler.Scheduler(
            graph=self.graph.dependency_dict,
            conditions=conditions,
            termination_conds=termination_conds,
        )

        # We also need to get the output nodes
        self.output_nodes = [
            n
            for n in self.graph.nodes
            if n.id not in (e.sender for e in self.graph.edges)
        ]

        # Lets also get the corresponding EvaluableNode for outputs
        self.output_enodes = [self.enodes[n.id] for n in self.output_nodes]

    def evaluate(
        self,
        time_increment: Union[int, float] = None,
        array_format: str = FORMAT_DEFAULT,
        initializer: Optional[Dict[str, Any]] = None,
    ):
        """
        Evaluates a :class:`~modeci_mdf.mdf.Graph`. This is the top-level interface to the execution engine.

        Args:
            time_increment: Time step for next execution
            array_format: A n-dimensional array
            initializer: sets the initial value of parameters of the node

        """
        # Any values that are set via the passed in initalizer, set their values. This lets us avoid creating
        # dummy input nodes with parameters for evaluating the graph
        for en_id, en in self.enodes.items():
            for inp_name, inp in en.evaluable_inputs.items():
                if initializer and inp_name in initializer:
                    inp.set_input_value(initializer[inp_name])

        print(
            "Evaluating graph: %s, root nodes: %s, with array format %s"
            % (self.graph.id, self.root_nodes, array_format)
        )
        str_conds_nb = "\n  ".join(
            [
                f"{node.id}: {cond}"
                for node, cond in self.scheduler.conditions.conditions.items()
            ]
        )
        str_conds_term = "\n  ".join(
            [
                f"{scale}: {cond}"
                for scale, cond in self.scheduler.termination_conds.items()
            ]
        )
        if self.verbose:
            print(" node-based conditions\n  %s" % str_conds_nb)
            print(" termination conditions\n  %s" % str_conds_term)

        incoming_edges = {n: set() for n in self.graph.nodes}
        for edge in self.graph.edges:
            incoming_edges[self.graph.get_node(edge.receiver)].add(edge)

        for ts in self.scheduler.run():
            if self.verbose:
                print(
                    "> Evaluating time step: %s"
                    % self.scheduler.get_clock(None).simple_time
                )
            for node in ts:
                for edge in incoming_edges[node]:
                    self.evaluate_edge(
                        edge, time_increment=time_increment, array_format=array_format
                    )
                self.enodes[node.id].evaluate(
                    time_increment=time_increment, array_format=array_format
                )

        if self.verbose:
            print("Order of execution of nodes\n")
            print(list(self.scheduler.run()))
            print("\n Trial terminated")

    def evaluate_edge(
        self,
        edge: Edge,
        time_increment: Union[int, float] = None,
        array_format: str = FORMAT_DEFAULT,
    ):
        """Evaluates edges in graph

        Args:
            time_increment: Time step for next execution
            array_format: A n-dimensional array

        """
        pre_node = self.enodes[edge.sender]
        post_node = self.enodes[edge.receiver]
        value = pre_node.evaluable_outputs[edge.sender_port].curr_value
        weight = (
            1
            if not edge.parameters or not "weight" in edge.parameters
            else edge.parameters["weight"]
        )

        if self.verbose:
            print(
                "  Edge %s connects %s to %s, passing %s with weight %s"
                % (
                    edge.id,
                    pre_node.node.id,
                    post_node.node.id,
                    _val_info(value),
                    _val_info(weight),
                )
            )
        if (type(weight) == int or type(weight) == float) and weight == 1:
            input_value = value
        else:
            input_value = weight * value
        post_node.evaluable_inputs[edge.receiver_port].set_input_value(input_value)

    def parse_condition(
        self, condition: Union[Condition, Dict]
    ) -> graph_scheduler.Condition:
        """Convert the condition in a specific format

        Args:
            condition: Specify the condition under which a Component should be allowed to execute

        Returns:
            Condition in specific format

        """

        def get_custom_parameter_getter(eobj):
            def getter(dependency, parameter):
                res = eobj.curr_value
                if res is None:
                    return 0
                else:
                    return res

            return getter

        def update_condition_arguments(args, condition_type):
            # mdf format prefers the key name for all conditions that use it or
            # its value as an __init__ argument
            combined_condition_arguments = {"dependencies": "dependency"}
            sig = inspect.signature(condition_type)

            for preferred, actual in combined_condition_arguments.items():
                if (
                    preferred in condition.kwargs
                    and preferred not in sig.parameters
                    and actual in sig.parameters
                ):
                    args[actual] = args[preferred]
                    del args[preferred]

            if "custom_parameter_getter" in sig.parameters:
                try:
                    dependency = args["dependency"]
                    parameter = args["parameter"]
                except KeyError as e:
                    raise ValueError(
                        "Threshold condition did not specify dependency or parameter"
                    ) from e

                # self is EvaluableGraph condition is running under,
                enode = self.enodes[dependency.id]
                evaluable_objects = [
                    enode.evaluable_inputs,
                    enode.evaluable_parameters,
                    enode.evaluable_functions,
                    enode.evaluable_outputs,
                ]
                valid_parameters = itertools.chain.from_iterable(
                    [obj for obj in evaluable_objects]
                )
                args["custom_parameter_validator"] = (
                    lambda dependency, parameter: parameter in valid_parameters
                )
                # assumes a unique parameter id among evaluable objects for
                # dependency, or will resolve in favor of first with that id
                for obj in evaluable_objects:
                    try:
                        obj = obj[parameter]
                    except KeyError:
                        pass
                    else:
                        args["custom_parameter_getter"] = get_custom_parameter_getter(
                            obj
                        )
                        break
                else:
                    raise ValueError(
                        f"No {parameter} evaluable object for {dependency}, options: {list(valid_parameters)}"
                    )

            return args

        # if specified as dict
        try:
            args = condition["kwargs"]
        except (TypeError, KeyError):
            args = {}

        try:
            condition = Condition(condition["type"], **args)
        except (TypeError, KeyError):
            pass

        cond_type = condition.type
        cond_args = copy.copy(condition.kwargs)

        try:
            typ = getattr(graph_scheduler.condition, cond_type)
        except AttributeError as e:
            raise ValueError("Unsupported condition type: %s" % cond_type) from e
        except TypeError as e:
            raise TypeError("Invalid condition dictionary: %s" % condition) from e

        for k, v in cond_args.items():
            new_v = self.graph.get_node(v)
            if new_v is not None:
                # arg is a node id
                cond_args[k] = new_v

            try:
                if isinstance(v, list):
                    # arg is a list all of conditions
                    new_v = [self.parse_condition(item) for item in v]
                else:
                    # arg is another condition
                    new_v = self.parse_condition(v)
            except (AttributeError, TypeError, ValueError):
                try:
                    # value may be a string representing a TimeScale
                    cond_args[k] = getattr(
                        graph_scheduler.TimeScale,
                        re.match(time_scale_str_regex, v).groups()[1],
                    )
                except (AttributeError, IndexError, TypeError):
                    pass
            else:
                cond_args[k] = new_v

        cond_args = update_condition_arguments(cond_args, typ)

        try:
            return typ(**cond_args)
        except TypeError as e:
            sig = inspect.signature(typ)

            try:
                var_positional_arg = [
                    name
                    for name, param in sig.parameters.items()
                    if param.kind is inspect.Parameter.VAR_POSITIONAL
                ][0]
            except IndexError:
                # other unhandled situation
                raise e
            else:
                try:
                    cond_args[var_positional_arg]
                except KeyError:
                    # error is due to missing required parameter,
                    # not named variable positional argument
                    raise TypeError(f"Condition '{typ.__name__}': {e}")
                else:
                    return typ(
                        *cond_args[var_positional_arg],
                        **{
                            k: v
                            for k, v in cond_args.items()
                            if k != var_positional_arg
                        },
                    )


def main(example_file: str, array_format: str = FORMAT_NUMPY, verbose: bool = False):
    """
    Main entry point for execution engine.

    Args:
        example_file: The MDF file to execute.
        array_format: The format of arrays to use. Allowed values: 'numpy' or 'tensorflow'.
        verbose: Whether to print output to standard out during execution.

    """

    from modeci_mdf.utils import load_mdf, print_summary

    mod_graph = load_mdf(example_file).graphs[0]

    if verbose:
        print("Loaded Graph:")
        print_summary(mod_graph)

        print("------------------")
    eg = EvaluableGraph(mod_graph, verbose)
    eg.evaluate(array_format=array_format)

    return eg


if __name__ == "__main__":

    example_file = os.path.join(
        os.path.dirname(__file__), "..", "..", "examples/MDF/Simple.json"
    )
    verbose = True
    if len(sys.argv) >= 2:
        example_file = sys.argv[1]

    if "-v" in sys.argv:
        verbose = True
    else:
        verbose = False

    from modelspec.utils import FORMAT_NUMPY, FORMAT_TENSORFLOW

    format = FORMAT_TENSORFLOW if "-tf" in sys.argv else FORMAT_NUMPY

    print("Executing MDF file %s with scheduler" % example_file)

    main(example_file, array_format=format, verbose=verbose)



"""
Implementations of importers and exporters for supported environments; fulfilling the
`hub and spoke model <https://github.com/ModECI/MDF/tree/main/examples>`_ of MDF by
allowing exchange between different modeling environments via MDF.

Most of these exporters and importers are currently a work and progress.
"""



"""
Code for exporting MDF models to ONNX.
"""

import re

from modeci_mdf.utils import load_mdf
from modeci_mdf.execution_engine import EvaluableGraph

import onnx
from onnx import helper, shape_inference
from onnx import AttributeProto, TensorProto, GraphProto
from onnx.defs import get_schema

from ast import literal_eval

import argparse
import os

import numpy


def mdf_to_onnx(mdf_model):
    """
    Takes an MDF model object and returns a list of ONNX models for each graph in the model.
    """

    # An MDF model can have multiple graphs. Each graph will be an onnx model
    onnx_models = []
    for graph in mdf_model.graphs:
        print("Processing Graph ", graph.id)

        # Use edges and nodes to construct execution order
        nodenames_in_execution_order = []
        evaluable_graph = EvaluableGraph(graph, verbose=False)
        for idx, edge in enumerate(evaluable_graph.ordered_edges):
            if idx == 0:
                nodenames_in_execution_order.append(edge.sender)
            nodenames_in_execution_order.append(edge.receiver)

        # print(nodenames_in_execution_order, graph.nodes, graph.edges)

        # Generate onnx graph
        onnx_graph = generate_onnx_graph(graph, nodenames_in_execution_order)

        # Make an onnx model from graph
        onnx_model = helper.make_model(onnx_graph)

        # Infer shapes
        onnx_model = shape_inference.infer_shapes(onnx_model)

        # Check model
        onnx.checker.check_model(onnx_model)

        onnx_models.append(onnx_model)

    return onnx_models


def generate_onnx_graph(graph, nodenames_in_execution_order):
    print("Generating ONNX graph for ", graph.id)

    onnx_graph_inputs = []
    onnx_graph_outputs = []
    onnx_nodes = []
    onnx_initializer = []

    for nodename in nodenames_in_execution_order:
        node = graph.get_node(nodename)

        # Get the node, and graph inputs, outputs from that node and the node's initializer
        (
            onnx_node,
            onnx_graph_input,
            onnx_graph_output,
            node_initializer,
        ) = generate_onnx_node(node, graph)

        onnx_nodes.append(onnx_node)
        onnx_graph_inputs.extend(onnx_graph_input)
        onnx_graph_outputs.extend(onnx_graph_output)
        onnx_initializer.extend(node_initializer)

    # Make the final graph
    onnx_graph = helper.make_graph(
        onnx_nodes, graph.id, onnx_graph_inputs, onnx_graph_outputs, onnx_initializer
    )

    return onnx_graph


def generate_onnx_node(node, graph):
    """
    Convert an MDF node into an ONNX node.
    Takes an MDF node, MDF graph and returns the ONNX node, any inputs to the node coming from  outside the graph,
    any outputs from the node going outside the graph, and an initializer for constants
    """

    onnx_graph_inputs = []
    onnx_graph_outputs = []
    onnx_initializer = []

    sender_port_name = {}  # Names of ports that send values to this node

    # Go over all parameters
    # Assumption: there may be multiple parameters but there will only be one function.
    # If there are multiple functions, the code should change so that each function should become its own node.
    for param in node.parameters:
        # If this is a constant
        if type(param.value) == int or type(param.value) == float:
            # Create a constant onnx node
            name = node.id + "_" + param.id
            constant = helper.make_tensor(
                name, data_type=TensorProto.FLOAT, dims=[], vals=[param.value]
            )
            onnx_initializer.append(constant)
            # The following will be the sender port from the constant onnx node for this parameter
            sender_port_name[param.id] = name
        elif param.function:
            # This is a function and will be part of an onnx node corresponding to this MDF node
            function_name = param.function

            onnx_function_prefix = "onnx::"
            pattern = re.compile(onnx_function_prefix)
            if re.match(pattern, function_name):
                # This is an onnx function
                function_name = function_name[len(onnx_function_prefix) :]
                # Get the arguments that this onnx function expects
                schema = get_schema(function_name)
                # The MDF description would have specified all the expected arguments of ths function
                function_input_names = [param.args[arg.name] for arg in schema.inputs]
            else:
                # Error
                raise "Cannot generate onnx function for the unknown function: {} specfied in the MDF node {}".format(
                    function_name,
                    node.id,
                )

    # Find the inputs to the new ONNX node. These are the senders of the in edges to this node
    node_in_edges = [edge for edge in graph.edges if edge.receiver == node.id]
    for in_edge in node_in_edges:
        sender_port_name[in_edge.receiver_port] = (
            in_edge.sender + "_" + in_edge.sender_port
        )
    onnx_node_input_names = [
        sender_port_name[function_input_name]
        if function_input_name in sender_port_name
        else function_input_name
        for function_input_name in function_input_names
    ]
    # No parameters. Constants became their own nodes earlier
    onnx_node_parameters = {}

    # Find the outputs of the new ONNX node. These are the output ports of the node
    onnx_node_output_names = [node.id + "_" + port.id for port in node.output_ports]

    # print(node.id, node_in_edges,node_out_edges)
    # print(function_name, onnx_node_input_names, onnx_node_output_names)

    # Create an ONNX node
    onnx_node = helper.make_node(
        function_name,
        onnx_node_input_names,
        onnx_node_output_names,
        name=node.id,
        **onnx_node_parameters,
    )

    # Check if any of the node's inputs are the inputs to the ONNX graph itself.
    # These are the node's inputs that don't have an incoming edge.
    input_ports_with_edge = [in_edge.receiver_port for in_edge in node_in_edges]
    input_ports_without_edge = [
        input_port
        for input_port in node.input_ports
        if input_port.id not in input_ports_with_edge
    ]
    if input_ports_without_edge:
        # Create ONNX graph input ports
        for input_port in input_ports_without_edge:
            value_info = helper.make_tensor_value_info(
                input_port.id, TensorProto.FLOAT, input_port.shape
            )
            onnx_graph_inputs.append(value_info)

    # Check if any of the node's outputs are the outputs of the ONNX graph.
    # These are the node's outputs that don't have an outgoing edge
    node_out_edges = [edge for edge in graph.edges if edge.sender == node.id]

    output_ports_with_edge = [out_edge.sender_port for out_edge in node_out_edges]
    output_ports_without_edge = [
        output_port
        for output_port in node.output_ports
        if output_port.id not in output_ports_with_edge
    ]
    if output_ports_without_edge:
        # Create ONNX graph output ports
        for output_port in output_ports_without_edge:
            # No need to create output shapes because they are inferred by ONNX
            value_info = helper.make_empty_tensor_value_info(
                node.id + "_" + output_port.id
            )
            onnx_graph_outputs.append(value_info)
    # print("Graph ip op", input_ports_without_edge, output_ports_without_edge)

    return onnx_node, onnx_graph_inputs, onnx_graph_outputs, onnx_initializer


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Converter from MDF to ONNX. "
        "Takes in a JSON/YAML file and generates ONNX files in the same directory as the input file"
    )
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_file",
        type=str,
        help="An input JSON/YAML file. "
        "Output files are generated in same directory "
        "with a -m2o.onnx extension",
    )

    args = parser.parse_args()

    convert_mdf_file_to_onnx(args.input_file)


def convert_mdf_file_to_onnx(input_file: str):
    """
    Converter from MDF to ONNX. Takes in a JSON/ONNX file and generates ONNX files.

    Args:
        input_file: The input file path to the MDF file. Output files are generated in same
            directory with -m2o.onnx extensions.

    Returns:
        NoneType
    """

    import os

    # Load the MDF model from file - this is not used anymore
    mdf_model = load_mdf(input_file)
    # print("Loaded MDF file from ", mdf_model_file)

    onnx_models = mdf_to_onnx(mdf_model)

    for onnx_model in onnx_models:
        out_filename = (
            f"{os.path.splitext(input_file)[0]}_{onnx_model.graph.name}-m2o.onnx"
        )
        onnx.save(onnx_model, out_filename)
        print("ONNX output saved in ", out_filename)


# Standalone execution
if __name__ == "__main__":
    main()



"""
Wrap commonly-used torch builtins in nn.Module subclass
for easier automatic construction of script
"""
import torch
import torch.nn as nn
import torch.nn.functional as F


class argmax(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):
        return torch.argmax(A)


class argmin(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):
        return torch.argmin(A)


class matmul(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A, B):
        return torch.matmul(A, B.T)


class add(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A, B):
        return torch.add(A, B)


class sin(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):
        return torch.sin(A)


class cos(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):
        return torch.cos(A)


class abs(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):
        return torch.abs(A)


class flatten(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):
        return torch.reshape(A, (1, -1))


class clip(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A, min_val, max_val):
        return torch.clamp(A, min_val, max_val)


class shape(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):
        return torch.tensor(A.size()).to(torch.int64)


class det(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):
        return torch.det(A)


class And(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A, B):
        return torch.logical_and(A > 0, B > 0)


class Or(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A, B):
        return torch.logical_or(A > 0, B > 0)


class Xor(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A, B):
        return torch.logical_xor(A > 0, B > 0)


class concat(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A, axis=0):

        return torch.cat(A, axis)


class ceil(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):

        return torch.ceil(A)


class floor(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, A):

        return torch.floor(A)


class bitshift(torch.nn.Module):
    def __init__(self, DIR):
        super().__init__()
        self.dir = DIR

    def forward(self, A, B):
        if self.dir == "RIGHT":
            return A.to(torch.int64) >> B.to(torch.int64)
        else:
            return A.to(torch.int64) << B.to(torch.int64)


class conv(torch.nn.Module):
    def __init__(
        self,
        auto_pad="NOTSET",
        kernel_shape=None,
        group=1,
        strides=[1, 1],
        dilations=[1, 1],
        pads=[0, 0, 0, 0],
    ):
        super().__init__()
        self.group = group
        self.auto_pad = auto_pad
        self.strides = tuple(strides)
        self.dilations = tuple(dilations)
        self.kernel_shape = kernel_shape

    def forward(self, A, W, B=None):
        if self.auto_pad == "NOTSET":
            self.pads = tuple(pads)

        elif self.auto_pad == "VALID":
            self.pads = (0, 0, 0, 0)
        elif self.auto_pad == "SAME_UPPER":
            pad_dim1 = (
                torch.ceil(torch.tensor(A.shape[2]).to(torch.float32) / strides[0])
                .to(torch.int64)
                .item()
            )
            pad_dim2 = (
                torch.ceil(torch.tensor(A.shape[3]).to(torch.float32) / strides[1])
                .to(torch.int64)
                .item()
            )
            if pad_dim1 % 2 == 0 and pad_dim2 % 2 == 0:
                self.pads = (pad_dim1 // 2, pad_dim1 // 2, pad_dim2 // 2, pad_dim2 // 2)
            elif pad_dim1 % 2 == 0 and pad_dim2 % 2 != 0:
                self.pads = (
                    pad_dim1 // 2,
                    pad_dim1 // 2,
                    pad_dim2 // 2,
                    pad_dim2 // 2 + 1,
                )
            elif pad_dim1 % 2 != 0 and pad_dim2 % 2 == 0:
                self.pads = (
                    pad_dim1 // 2,
                    pad_dim1 // 2 + 1,
                    pad_dim2 // 2,
                    pad_dim2 // 2,
                )
            elif pad_dim1 % 2 != 0 and pad_dim2 % 2 != 0:
                self.pads = (
                    pad_dim1 // 2,
                    pad_dim1 // 2 + 1,
                    pad_dim2 // 2,
                    pad_dim2 // 2 + 1,
                )

        elif self.auto_pad == "SAME_LOWER":

            pad_dim1 = (
                torch.ceil(torch.tensor(A.shape[2]).to(torch.float32) / strides[0])
                .to(torch.int64)
                .item()
            )
            pad_dim2 = (
                torch.ceil(torch.tensor(A.shape[3]).to(torch.float32) / strides[1])
                .to(torch.int64)
                .item()
            )
            if pad_dim1 % 2 == 0 and pad_dim2 % 2 == 0:
                self.pads = (pad_dim1 // 2, pad_dim1 // 2, pad_dim2 // 2, pad_dim2 // 2)
            elif pad_dim1 % 2 == 0 and pad_dim2 % 2 != 0:
                self.pads = (
                    pad_dim1 // 2,
                    pad_dim1 // 2,
                    pad_dim2 // 2 + 1,
                    pad_dim2 // 2,
                )
            elif pad_dim1 % 2 != 0 and pad_dim2 % 2 == 0:
                self.pads = (
                    pad_dim1 // 2 + 1,
                    pad_dim1 // 2,
                    pad_dim2 // 2,
                    pad_dim2 / 2,
                )
            elif pad_dim1 % 2 != 0 and pad_dim2 % 2 != 0:
                self.pads = (
                    pad_dim1 // 2 + 1,
                    pad_dim1 // 2,
                    pad_dim2 // 2 + 1,
                    pad_dim2 // 2,
                )

        A = F.pad(A, self.pads)
        return F.conv2d(
            A,
            W,
            bias=B,
            stride=self.strides,
            padding=self.pads,
            dilation=self.dilations,
            groups=self.group,
        )


class elu(torch.nn.Module):
    def __init__(self, alpha=1.0):

        super().__init__()
        self.alpha = alpha

    def forward(self, A):

        return nn.ELU(alpha=self.alpha)(A.to(torch.float32))


class hardsigmoid(torch.nn.Module):
    def __init__(self, alpha=0.2, beta=0.5):

        super().__init__()
        self.alpha = alpha
        self.beta = beta

    def forward(self, A):

        return torch.clamp(self.alpha * (A.to(torch.float32)) + self.beta, 0, 1)


class hardswish(torch.nn.Module):
    def __init__(self):

        super().__init__()
        self.alpha = 1.0 / 6
        self.beta = 0.5

    def forward(self, A):

        return A * torch.clamp(self.alpha * (A.to(torch.float32)) + self.beta, 0, 1)


class hardmax(torch.nn.Module):
    def __init__(self, axis=-1):
        super().__init__()

        self.axis = axis

    def forward(self, A):
        A = A.to(torch.float32)
        rank = A.shape
        if self.axis < 0:
            self.axis += len(rank)
        tensor = torch.arange(rank[self.axis])
        repeats = []
        repeats.append(1)
        for i, idx in enumerate(reversed(rank[: self.axis])):

            repeats.append(1)
            tensor = torch.stack([tensor] * idx)

        for i, idx in enumerate(rank[self.axis + 1 :]):
            repeats.append(idx)

            tensor = tensor.unsqueeze(-1).repeat(repeats)
            repeats[-1] = 1
        # b = torch.stack([torch.stack([torch.arange(4)] * 3)] *2)
        # print(tensor.shape)
        max_values, _ = torch.max(A, dim=self.axis)
        # print(max_values, max_values.shape)
        # tensor = torch.reshape(tensor, tuple(rank))
        tensor[A != torch.unsqueeze(max_values, dim=self.axis)] = rank[self.axis]
        # print(b)
        first_max, _ = torch.min(tensor, dim=self.axis)

        one_hot = torch.nn.functional.one_hot(first_max, rank[self.axis])
        return one_hot


class compress(torch.nn.Module):
    def __init__(self, axis=None):
        self.axis = axis
        super().__init__()

    def forward(self, A, B):

        idx = (B.to(torch.bool) != 0).nonzero().reshape(-1)
        if self.axis != None:
            return torch.index_select(A, self.axis, idx)

        else:

            return torch.index_select(A.reshape(-1), 0, idx)


# TODO: Many more to be implemented


__all__ = [
    "argmax",
    "argmin",
    "matmul",
    "add",
    "sin",
    "cos",
    "abs",
    "flatten",
    "clip",
    "shape",
    "det",
    "And",
    "Or",
    "Xor",
    "concat",
    "ceil",
    "floor",
    "bitshift",
    "conv",
    "elu",
    "hardsigmoid",
    "hardswish",
    "compress",
]



"""
MDF is intended to be an open source, community-supported standard and associated library of tools for expressing
computational models in a form that allows them to be exchanged between diverse programming languages and execution
environments. The MDF Python API can be used to create or load an MDF model for inspection and validation. It also
includes a basic execution engine for simulating models in the format. However, this is not intended as a general
purpose simulation environment, nor is MDF intended as a programming language.  Rather, the primary purpose of the
Python API is to facilitate and validate the exchange of models between existing environments that serve different
communities. Accordingly, these Python tools include bi-directional support for importing to and exporting from
widely-used programming environments in a range of disciplines, and for easily extending these to other environments.
"""

# Version of the specification for MDF
MODECI_MDF_VERSION = "0.4"

# Version of the Python module.
__version__ = "0.4.1.1"



"""
Specifies and implements the MDF the function ontology; a collection of builtin functions that can be used in
MDF :class:`~modeci.mdf.Function` and :class:`~modeci.mdf.Parameter` objects. Code for registering standard functions
are with the ontology is implemented in :mod:`~modeci_mdf.functions.standard`. If you want to add functions to the
ontology that can be used during execution see how this has been done for ONNX (:mod:`~modeci_mdf.functions.onnx`) and
ACT-R (:mod:`~modeci_mdf.functions.actr`)
"""



import re

basestring = str


class PatternException(Exception):
    pass


def get(obj, name, key):
    if name is None:
        a = obj
    else:
        a = obj[name]

    while type(key) == str and "." in key:
        key1, key = key.split(".", 1)
        try:
            a = a[key1]
        except AttributeError:
            a = getattr(a, key1)

    try:
        x = a[key]
    except AttributeError:
        x = getattr(a, key)
    if isinstance(x, float):
        x = "%g" % x
    if not isinstance(x, str):
        x = repr(x)
    return x


def partialmatch(obj, name, key, b, value):
    if type(key) == str and key[0] == "?":
        key = b[key[1:]]
    v = get(obj, name, key)
    if v == value:
        return True

    # fix for early Python versions where True and False are actually 1 and 0
    if value in ["True", "False"] and type(True) == int:
        if v == str(bool(value)):
            return True

    pm = b.get("_partial", None)
    if pm is not None:
        x = pm.match(key, value, v)
        obj._partial += x
        return True
    else:
        return False


class Pattern:
    def __init__(self, patterns, bound=None, partial=None):
        self.funcs = parse(patterns, bound)
        self.partial = partial

    def match(self, obj):
        b = {}
        b["_partial"] = self.partial
        if self.partial is not None:
            obj._partial = 0.0
        try:
            for f in self.funcs:
                if f(obj, b) == False:
                    return None
        except (AttributeError, TypeError, KeyError):
            return None
        del b["_partial"]
        return b


def parse(patterns, bound=None):
    if not hasattr(patterns, "items"):
        patterns = {None: patterns}
    funcs = []
    vars = {}
    funcs2 = []
    for name, pattern in patterns.items():
        if not isinstance(pattern, (list, tuple)):
            pattern = [pattern]
        for p in pattern:
            if p is None:
                if name is None:
                    funcs.append(lambda x, b: x == None)
                else:
                    funcs.append(
                        lambda x, b, name=name: x[name] == None or len(x[name]) == 0
                    )
            elif callable(p):
                if name is None:

                    def callfunc(x, b, name=name, p=p):
                        return p(x, b)

                else:

                    def callfunc(x, b, name=name, p=p):
                        return p(x[name], b)

                funcs2.append(callfunc)
            elif isinstance(p, basestring):
                namedSlots = False
                for j, text in enumerate(p.split()):
                    key = j
                    m = re.match(r"([?]?[\w\.]+):", text)
                    if m != None:
                        key = m.group(1)
                        try:
                            key = int(key)
                        except ValueError:
                            pass
                        text = text[m.end() :]
                        if len(text) == 0:
                            raise PatternException(
                                "No value for slot '%s' in pattern '%s'"
                                % (key, pattern)
                            )
                        namedSlots = True
                    else:
                        if namedSlots != False:
                            raise PatternException(
                                "Found unnamed slot '%s' after named slot in pattern '%s'"
                                % (text, pattern)
                            )
                    if text == "?":
                        continue
                    while len(text) > 0:
                        m = re.match(r"([\w\.-]+)", text)
                        if m != None:
                            text = text[m.end() :]
                            t = m.group(1)
                            funcs.append(
                                lambda x, b, name=name, key=key, t=t: partialmatch(
                                    x, name, key, b, t
                                )
                            )
                            continue

                        m = re.match(r"!([\w\.-]+)", text)
                        if m != None:
                            text = text[m.end() :]
                            t = m.group(1)
                            funcs.append(
                                lambda x, b, name=name, key=key, t=t: get(x, name, key)
                                != t
                            )
                            continue

                        m = re.match(r"\?(\w+)", text)
                        if m != None:
                            text = text[m.end() :]
                            v = m.group(1)
                            if bound is not None and v in bound:
                                funcs.append(
                                    lambda x, b, name=name, key=key, t=bound[
                                        v
                                    ]: partialmatch(x, name, key, b, t)
                                )
                            elif v in vars:
                                funcs2.append(
                                    lambda x, b, name=name, key=key, v=v: partialmatch(
                                        x, name, key, b, b[v]
                                    )
                                )
                            else:
                                vars[v] = (name, key)

                                def setfunc(x, b, name=name, key=key, v=v):
                                    b[v] = get(x, name, key)
                                    return True

                                funcs.append(setfunc)
                            continue

                        m = re.match(r"!\?(\w+)", text)
                        if m != None:
                            text = text[m.end() :]
                            v = m.group(1)
                            if bound is not None and v in bound:
                                funcs.append(
                                    lambda x, b, name=name, key=key, t=bound[v]: get(
                                        x, name, key
                                    )
                                    != t
                                )
                            else:
                                funcs2.append(
                                    lambda x, b, name=name, key=key, v=v: get(
                                        x, name, key
                                    )
                                    != b[v]
                                )
                            continue

                        raise PatternException(
                            f"Unknown text '{text}' in pattern '{pattern}'"
                        )
    return funcs + funcs2



import atexit
import time
import bisect
import sys
import random
import os

using_java = False
try:
    from java.io import File

    using_java = True
except:
    pass


def file_exists(filename):
    if hasattr(os, "access"):
        return os.access(filename, os.F_OK)
    elif using_java:
        return File(filename).exists()


try:
    basestring  # For Python 2 compatibility
except NameError:
    basestring = str  # For Python 3 compatibility


class Trace:
    def __init__(self):
        self.data = dict(time=[(0, 0.0)])
        self.index = 1
        self.type = {}

    def add(self, key, value):
        if not isinstance(value, (bool, int, float, basestring, type(None))):
            value = repr(value)
        if type(value) == str:
            if value.startswith("<"):
                return
        if not self.data.has_key(key):
            self.data[key] = []

        self.data[key].append((self.index, value))
        self.index += 1

    def keys(self):
        return self.data.keys()

    def fixed_keys(self):
        return [k for k, v in self.data.items() if len(v) == 1]

    def get_final(self, key):
        return self.data[key][-1][1]

    def get_pts(self, vars):
        pts = []
        for v in vars:
            for i, val in self.data[v]:
                if i not in pts:
                    pts.append(i)
        pts.sort()
        return pts

    def merge_pts(self, pts, key):
        times = [x[0] for x in self.data[key]]
        i = 0
        for t in times:
            if i >= len(pts) or t < pts[i]:
                continue
            while i < len(pts) - 1 and pts[i + 1] < t:
                del pts[i]
            i += 1
        while i < len(pts) - 1:
            del pts[i]

    def group_pts(self, pts, key):
        times = [x[0] for x in self.data[key]]
        group = []
        i = 0
        for p in pts:
            if i >= len(times) or p < times[i]:
                group.append(p)
            else:
                if len(group) > 0:
                    yield group
                group = [p]
                while i < len(times) and p >= times[i]:
                    i += 1
        if len(group) > 0:
            yield group

    def get_at(self, name, time):
        d = self.data[name]
        i = bisect.bisect(d, (time + 0.5,)) - 1
        if i < 0:
            return None
        if i >= len(d) - 1:
            i = -1
        return d[i][1]

    def __nonzero__(self):
        return self.index != 1


class Log:
    def __init__(self):
        self.do_screen = True
        self.do_html = False
        self.do_summary = False
        self.do_data = False
        self.directory = sys.argv[0]
        self.using_default_directory = True
        if self.directory.endswith(".py"):
            self.directory = self.directory[:-3]

        self.start_time = time.time()
        self.last_flush = self.start_time

        self.reset()

    def use_directory(self, dir):
        self.directory = dir
        self.using_default_directory = False

    def set(self, key, value):
        if key == "time":
            self.time = value
        if self.do_screen:
            self.display_value(key, value)
        if self.do_data or self.do_summary:
            self.data[key] = value
        if self.do_html:
            self.trace.add(key, value)

    def display_value(self, key, value):
        if key != "time":
            print(f"{self.time:8.3f} {key} {value}")

    def display_all(self):
        if self.time > 0:
            print("Total time: %8.3f" % self.time)
        items = self.data.items()
        items.sort()
        for k, v in items:
            if k != "time":
                print(f" {k} {v}")

    def get_time_code(self):
        t = time.strftime("%Y%m%d-%H%M%S", time.localtime(self.start_time))
        return f"{t}-{self.id}"

    def ensure_directory_exists(self):
        if not file_exists(self.directory + "/"):
            os.makedirs(self.directory)

    def reset(self):
        self.trace = Trace()
        self.data = {}
        self.time = 0
        self.id = "%08x" % int(random.randrange(0x7FFFFFFF))


class DummyLog:
    def set(self, key, value):
        pass

    def __nonzero__(self):
        return False

    def __setattr__(self, key, value):
        pass

    def __getattr__(self, key):
        if key[0] != "_":
            return self
        else:
            raise AttributeError(key)

    def __setitem__(self, key, value):
        pass

    def __getitem__(self, key):
        return self


dummy = DummyLog()


class LogProxy:
    def __init__(self, log, prefix=""):
        self._log = log
        if prefix.startswith("."):
            prefix = prefix[1:]
        self._prefix = prefix
        self._sub = {}

    def __setattr__(self, key, value):
        if key.startswith("_"):
            if key == "_":
                self._set(value)
            else:
                self.__dict__[key] = value
        else:
            getattr(self, key)._set(value)

    def __getattr__(self, key):
        if key.startswith("_"):
            if key == "_":
                return self
            if key in self.__dict__:
                return self.__dict__[key]
            raise AttributeError(key)
        else:
            s = LogProxy(self._log, f"{self._prefix}.{key}")
            self.__dict__[key] = s
            return s

    def __setitem__(self, key, value):
        self[key]._set(value)

    def __getitem__(self, key):
        s = self._sub.get(key, None)
        if s is not None:
            return s
        s = LogProxy(self._log, f"{self._prefix}[{key}]")
        self._sub[key] = s
        return s

    def _set(self, value):
        try:
            if not isinstance(value, (int, float, bool, basestring, tuple, list, dict)):
                value = repr(value)
        except TypeError:
            value = repr(value)
        self._log.set(self._prefix, value)

    def __nonzero__(self):
        return True


singleton_log = Log()
log_proxy = LogProxy(singleton_log)


def log(screen=None, html=None, data=None, summary=None, directory=None):
    if screen is not None:
        singleton_log.do_screen = screen
    if html is not None:
        singleton_log.do_html = html
    if data is not None:
        singleton_log.do_data = data
    if summary is not None:
        singleton_log.do_summary = summary
    if directory is not None:
        singleton_log.use_directory(directory)
    return log_proxy


pending_output = []


def finished(flush=True):
    log = singleton_log
    has_data = log.data or log.trace

    if has_data:
        if log.do_summary:
            log.display_all()
        if log.do_data:
            log.ensure_directory_exists()
            fn = f"{log.directory}/{log.get_time_code()}.data"
            pending_output.append((fn, log.data))
        if log.do_html:
            log.ensure_directory_exists()
            from ccm.ui.htmltrace import HTMLTrace

            html = HTMLTrace(log.trace)
            html.generate(f"{log.directory}/{log.get_time_code()}.html")

    if flush or time.time() - log.last_flush > 10:
        for fn, data in pending_output:
            f = file(fn, "w")
            items = data.items()
            items.sort()
            for k, v in items:
                f.write(f"{k}={v}\n")
            f.close()
        del pending_output[:]
        log.last_flush = time.time()
    log.reset()


atexit.register(finished, flush=True)



"""
Functions for converting from PyTorch TorchScript to MDF models.

This code was originally inspired by the following blog post:

    Mike He, "From Models to Computation Graphs (Part I)", https://ad1024.space/articles/22
"""
import inspect
import logging

from typing import Union, Dict, Any, Tuple, List, Callable
import onnx.defs


import torch

from modeci_mdf.mdf import Model, Graph, Node, Edge, InputPort, OutputPort, Parameter
from modeci_mdf.functions.onnx import onnx_opset_version as modeci_onnx_opset_version


logger = logging.getLogger(__name__)


def convert_to_serializable(value):
    """Helper function that converts some common unserializable types to JSON seriralizable types"""
    if type(value) is torch.device:
        value = str(value)
    elif type(value) is torch.Tensor:
        value = value.numpy().tolist()

    return value


def make_node_id(node: torch.Node) -> str:
    """Helper function to get a unique name (used in MDF as id) from a TorchScript Node object"""
    return "_".join(
        [node.kind().split("::")[-1]] + [str(o.unique()) for o in node.outputs()]
    )


def make_func_id(node: torch.Node) -> str:
    """Helper function to get a unique name (used in MDF as id) for a TorchScript node's op/function."""
    return f"{node.kind()}_1"


def make_model_graph_name(
    model: Union[torch.ScriptModule, torch.ScriptFunction]
) -> Tuple[str, str]:
    """Helper function that generates a clean graph and model name from a TorchScript model"""
    # Get a name for this module
    try:
        model_name = model.original_name.split(".")[-1]
        graph_name = f"{model_name}Graph"
    except AttributeError:
        try:
            model_name = model.qualified_name.split(".")[-1]
            graph_name = f"{model_name}_graph"
        except AttributeError:
            # It hasn't been compiled yet, use the class name I guess
            model_name = type(model).__name__.split(".")[-1]
            graph_name = f"{model_name}Graph"

    return model_name, graph_name


def process_torch_schema(
    node: torch.Node, consts: Dict, port_mapper: "PortMapper"
) -> Tuple[List[str], Dict[str, Any]]:
    """
    Parse a TorchScript node schema into argument names and constant attributes (parameters in MDF)

    Args:
        node: The TorchScript node to retrieve the schema from.
        consts: The constant nodes Dict for the graph we are working with.

    Returns:
        A tuple containing a list of argument names and Dict of parameter names and values.
    """

    # Get the input node names
    inputs = [i.unique() for i in node.inputs()]

    # If this is a TorchScript funciton (aten::*), it should have a schema string to parse.
    if "no schema" not in node.schema():
        schema = torch._C.parse_schema(node.schema())

        # Get the arguments and covert to a simple List[str]
        schema_args = schema.arguments
        schema_args = [schema_args[i].name for i, inp in enumerate(inputs)]

    else:
        logger.warning(
            f"Schema not found for TorchScript node ({node}), using placeholders for argument names."
        )
        schema_args = [f"arg{i}" for i in range(len(inputs))]

    # Get any input to this node that is TorchScript node.kind() prim::Constant, make those a parameter
    parameters = {
        schema_args[i]: consts[inp] for i, inp in enumerate(inputs) if inp in consts
    }

    return schema_args, parameters


def process_onnx_schema(
    node: torch.Node, port_mapper: "PortMapper"
) -> Tuple[Dict[str, str], Dict[str, Any]]:
    """
    Retrieve the argument names and attributes (parameters in MDF) for this Operation.

    Args:
        op: The TorchScript node containing the ONNX operation.
        port_mapper: The utitlity class for assigning TorchScript input output ids to Input Output Port ids.

    Returns:
        A two element tuple:
            - A dict representing argument names mapping to input port ids
            - A dict mapping parameters (ONNX attributes) names mapping to values
    """

    # Get the input node names
    inputs = [i.unique() for i in node.inputs()]

    # If this is an ONNX op, we need to get the schema from ONNX
    if "onnx::" in node.kind():
        try:
            schema = onnx.defs.get_schema(node.kind().split("::")[-1])

            schema_args = {}
            if len(schema.inputs) > 0:
                # If the first argument is variadic. Represent this as a list of input port names
                if schema.inputs[0].option.name == "Variadic":
                    schema_args = {
                        schema.inputs[0].name: str(
                            [
                                port_mapper.id_to_port(inp)
                                for i, inp in enumerate(inputs)
                            ]
                        )
                    }
                else:
                    schema_args = {
                        schema.inputs[i].name: port_mapper.id_to_port(inp)
                        for i, inp in enumerate(inputs)
                    }

        except onnx.onnx_cpp2py_export.defs.SchemaError:
            logger.warning(
                f"Could not find ONNX OpSchema for op {node.kind()}, using placeholder names for arguments."
            )
            schema_args = {
                f"arg{i}": port_mapper.id_to_port(inp) for i, inp in enumerate(inputs)
            }
    else:
        raise ValueError(f"Cannot process ONNX schema for non ONNX node: {node}")

    # ONNX attributes are equivalent to MDF parameters really
    parameters = {
        aname: convert_to_serializable(node[aname]) for aname in node.attributeNames()
    }

    return schema_args, parameters


def get_graph_constants(graph: torch.Graph) -> Dict[str, Any]:
    """
    Find all constant nodes in the graph and extract their values as a proper JSON serializable value.

    Args:
        graph: The graph to extract constants from.

    Returns:
        A Dict that maps the constant nodes unique TorchScript node ID string to its value.
    """
    consts = {}
    for n in graph.findAllNodes("prim::Constant"):
        for o in n.outputs():
            consts[o.unique()] = convert_to_serializable(o.toIValue())

    return consts


class PortMapper:
    r"""
    A simple class that handles mapping TorchScript input\ouput ids to MDF InputPort\OutputPort ids. It keeps track of
    annoying details like graph level inputs and stuff.
    """

    def __init__(self, graph: torch.Graph, args: Tuple):

        # Keep generate special names for all the graph inputs and parameters
        self.graph_inputs = PortMapper._get_graph_inputs_dict(graph, args)

    def id_to_port(self, id: str):
        """Turn unique TorchScript output and input value names into valid MDF input and outport names"""

        # If this id is a graph input, use its debug name
        if id in self.graph_inputs:
            id = self.graph_inputs[id]

        new_name = str(id).replace(".", "_")

        # Remove :: from ids, these cause issues with parsing in the execution engine
        new_name = new_name.replace("::", "_")

        # If the first character is a digit, precede with an underscore so this can never be interpreted
        # as number down the line.
        if new_name[0].isdigit():
            new_name = "_" + new_name

        return new_name

    def port_to_id(self, name: str):
        """Transform a port name back to is TorchScript ID"""

        # If first character is underscore, remove it
        id = name
        if name[0] == "_":
            id = name[1:]

        # Replace any remaining underscores with '.'
        id = id.replace("_", ".")

        # If this is a numeric id, make it an int again
        if id[0].isdigit():
            id = int(id)

        # If this id is actually a debugName from a graph input, use that
        for input_id, debug_name in self.graph_inputs.items():
            if debug_name == id:
                return input_id

        return id

    @staticmethod
    def _get_graph_inputs_dict(
        graph: torch.Graph, args: Tuple[torch.Tensor]
    ) -> Dict[str, str]:
        """
        Create a dict mapping graph input torch.Node ids to default names. The default names are just:
            - input1
            - input2
            - etc.

        Any parameters for the model will also be graph inputs but their node.debugName() will be used
        instead.
        """
        graph_inputs = {
            inp.unique(): inp.debugName() for i, inp in enumerate(graph.inputs())
        }

        # The first len(args) inputs should be the input arguments to the function or forward method. Lets
        # canonicalize them.
        input_ids = list(graph_inputs.keys())
        for i in range(len(args)):
            graph_inputs[input_ids[i]] = f"input{i + 1}"

        return graph_inputs


def torchnode_to_mdfnode(
    node: torch.Node,
    graph: torch.Graph,
    consts: Dict[str, Any],
    port_mapper: "PortMapper",
) -> Union[Node, None]:
    """
    Convert a TorchScript node to an MDF node.

    Args:
        node: The node to convert.
        graph: The graph that this node is a member.
        consts: A dict containing any constants in the graph.

    Returns:
        The MDF node for this TorchScript node. prim::Constant nodes are excluded from the MDF graph and are
        instead placed as parameters. In this case, return None.
    """
    op = node.kind()

    # Exclude constants (as nodes) from the MDF graph. We will instead insert them as parameters to the nodes that
    # they project to.
    if op == "prim::Constant":
        return None

    # If we are dealing with a loop node, we need to recursively create a sub-graph for the loop body
    if op == "onnx::Loop":
        sub_mdf_graph = Graph(id=f"LoopSubgraph{make_node_id(node)}")
        block_graph = list(node.blocks())[0]
        translate_graph(
            graph=block_graph,
            mdf_graph=sub_mdf_graph,
            consts=consts,
            port_mapper=port_mapper,
        )
        return sub_mdf_graph

    outputs = [o.unique() for o in node.outputs()]
    inputs = [i.unique() for i in node.inputs()]

    # Get the argument names and parameter names and values for this Node's operation
    if "onnx::" in op:
        arguments, parameters = process_onnx_schema(node, port_mapper)
    else:
        arguments, parameters = process_torch_schema(node, consts, port_mapper)

    mdf_node = Node(id=make_node_id(node))
    for p in parameters:
        mdf_node.parameters.append(Parameter(id=p, value=parameters[p]))

    # Add any output ports
    for o in outputs:
        mdf_node.output_ports.append(
            OutputPort(id=port_mapper.id_to_port(o), value=make_func_id(node))
        )

    # Add any input ports to the node, exclude inputs from constant nodes, these are parameters now
    for inp_i, inp in enumerate(inputs):
        if inp not in consts:
            ip_name = port_mapper.id_to_port(inp)

            # Try to get the shape and type of the input port
            inp_type = node.inputsAt(inp_i).type()
            try:
                shape = tuple(inp_type.sizes()) if inp_type.sizes() else None
            except RuntimeError:
                shape = None

            mdf_node.input_ports.append(
                InputPort(id=ip_name, shape=shape, type=str(inp_type))
            )

    # Add Parameter
    if type(arguments) == list:
        arguments = {"arguments": arguments}
    f = Parameter(id=make_func_id(node), function=op, args=arguments)
    mdf_node.parameters.append(f)

    return mdf_node


def translate_graph(
    graph: Union[torch.Graph, torch.Block],
    mdf_graph: Graph,
    consts: Dict[str, Any],
    port_mapper: "PortMapper",
):
    """
    Go through a :class:`~torch.Graph` or :class:`~torch.Block` and translate the nodes and edges to MDF nodes and
    edges.

    Args:
        graph: The graph to translate.
        mdf_graph: The MDF graph to store the translation into.
        consts: Constant to use for parameters of nodes.
        port_mapper: A port mapper instance to handle translating names.

    Returns:

    """

    # For every node, cache its input edges. This will let us look this up quickly for
    # any node in the loop below.
    node_to_in_edge = {
        node: [i.unique() for i in node.inputs()] for node in graph.nodes()
    }

    for node in graph.nodes():

        mdf_node = torchnode_to_mdfnode(
            node=node, graph=graph, consts=consts, port_mapper=port_mapper
        )

        # If we are excluding this node from the MDF graph, skip it.
        if mdf_node is None:
            continue

        mdf_graph.nodes.append(mdf_node)

        if type(mdf_node) == Graph:
            continue

        # Now we need to examine all outgoing edges from this node and add them to the MDF graph. We do this by looping
        # over all nodes in the graph and seeing if they have an input from the node we just constructed. This is
        # O(n^2) in terms of the number of the nodes!
        outputs = [o.unique() for o in node.outputs()]
        for to in graph.nodes():

            # Lookup this nodes input edges
            to_inputs = node_to_in_edge[to]

            edges = set(outputs) & set(to_inputs)
            for edge in edges:
                from_id = make_node_id(node)
                from_port = mdf_node.output_ports[outputs.index(edge)].id
                to_id = make_node_id(to)
                to_port = mdf_node.output_ports[outputs.index(edge)].id
                mdf_edge = Edge(
                    id=f"{from_id}_{to_id}",
                    sender=from_id,
                    sender_port=f"{from_port}",
                    receiver=to_id,
                    receiver_port=f"{to_port}",
                )
                mdf_graph.edges.append(mdf_edge)


def pytorch_to_mdf(
    model: Union[Callable, torch.nn.Module, torch.ScriptFunction, torch.ScriptModule],
    args: Union[None, torch.Tensor, Tuple[torch.Tensor]] = None,
    trace: bool = False,
    use_onnx_ops: bool = True,
) -> Union[Model, Graph]:
    r"""
    Convert a PyTorch model to an MDF model. By default, this function will invoke `torch.jit.script` on the
    model to compile it down to TorchScript IR and simplify the graph before exporting the MDF. The default is
    to use ONNX operations when possible and fallback to ATEN\Torch ops when ONNX support is not available
    (`torch._C._onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK` mode). To use allATEN\Torch ops, set use_onnx_ops to False.

    Args:
        model: The model to translate into MDF.
        args: The input arguments for this model. If a nn.Module is passed then the model will be traced with these
            inputs. If a ScriptModule is passed, they are still needed to deterimine input shapes.
        trace: Force the use of tracing to compile the model. The default is to use torch.jit.script
        use_onnx_ops: Use ONNX ops when possible, fallback to ATEN ops when not available. Default is True. If False,
            use only ATEN ops.

    Returns:
        The translated MDF model
    """

    # Get the graph and nodes from the TorchScript model
    try:
        # If the graph attribute is available, we are dealing with a already jitted model (ScriptModule, ScriptFunciton,
        # etc.)
        graph = model.graph
        jit_model = model
    except AttributeError:

        # Lets jit things, if the user doesn't want to trace or we are dealing with a standard Python function, we need
        # to JIT script it.
        if not trace or inspect.isfunction(model):
            jit_model = torch.jit.script(model)
            graph = jit_model.graph
        else:
            # If the user wants to trace, _model_to_graph below will take care of that for us.
            graph = None

    if use_onnx_ops:
        operator_export_type = torch._C._onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK
    else:
        operator_export_type = torch._C._onnx.OperatorExportTypes.RAW

    # Call out to a part of the ONNX exporter that simiplifies the graph before ONNX export.
    from torch.onnx.utils import _model_to_graph
    from torch.onnx import TrainingMode
    from torch.onnx.symbolic_helper import (
        _export_onnx_opset_version,
        _set_opset_version,
    )

    previous_opset_version = _export_onnx_opset_version
    _set_opset_version(modeci_onnx_opset_version)
    graph, params_dict, torch_out = _model_to_graph(
        model=jit_model if graph else model,
        args=args,
        do_constant_folding=True,
        training=TrainingMode.EVAL,
        operator_export_type=operator_export_type,
        dynamic_axes={},
    )
    _set_opset_version(previous_opset_version)

    model_name, graph_name = make_model_graph_name(model)

    # Setup the MDF model and graph
    mdf_model = Model(id=model_name)
    mdf_graph = Graph(id=graph_name)
    mdf_model.graphs.append(mdf_graph)

    # Get all constant nodes in the graph
    consts = get_graph_constants(graph)

    # Get any inputs to the graph, and their debug names. Pass args so we know how
    # many original input arguments the graph has. ONNX lowering from _model_to_graph
    # makes all parameters to the model inputs.
    port_mapper = PortMapper(graph=graph, args=args)

    # Translate the TorchScript graph to and MDF graph object. This could be a recursive call
    translate_graph(
        graph=graph, mdf_graph=mdf_graph, consts=consts, port_mapper=port_mapper
    )

    # Replace in "." for "_" in parameter names. We have done this elsewhere when creating the input ports for these
    # parameters.
    params_dict = {port_mapper.id_to_port(k): v for k, v in params_dict.items()}

    # Set the ONNX opset version
    mdf_model.onnx_opset_version = _export_onnx_opset_version

    return mdf_model, params_dict


if __name__ == "__main__":

    def simple(x, y):
        return x + y

    mdf_model, param_dict = pytorch_to_mdf(
        simple,
        args=(torch.tensor(1.0), torch.tensor(2.0)),
    )



"""
Functions for converting from MDF models to PyTorch
"""
import collections
import os
import sys
import h5py
import importlib
from collections import defaultdict
from inspect import getmembers, signature, getsource, isclass
import modeci_mdf
import numpy as np
import sympy
import torch
from typing import Union, Dict, Any, Tuple, List, Callable
import torch.nn as nn
from modeci_mdf.functions.standard import mdf_functions
from modeci_mdf.utils import load_mdf
from modeci_mdf.execution_engine import EvaluableGraph


not_self = set()
param_set = set()
graph_input = []


def get_module_declaration_text(
    name: str, node_dict: Dict[Any, Any], execution_order: List[str], version: str
):
    """
    Helper function to generate string in module.py. String will create an
    instance of the torch object corresponding to the node and function at node as method
    Generated text specifies the definition of class that will form the pytorch model,
    including __init__method and forward method.
    Returns string of the class definition with parameters and arguments assigned.

    Args:
        name: Name of the node
        node_dict: dictionary with attributes of the node such as Input Ports, functions, parameters and
        execution_order: List of nodes in the order of execution

    Returns: Script in PyTorch schema
    """
    declaration_text = ("\n\nclass {}(nn.Module):").format(name)

    input_ports = node_dict["input_ports"]
    functions = node_dict["functions"]
    parameters = node_dict["parameters"]
    output_ports = node_dict["output_ports"]
    if version == "mdf.s":
        function_set = set()
        for parameter in parameters:
            if parameter.function:
                function_set.add(parameter)

        torch.set_printoptions(profile="full")
        if parameters and not functions:
            declaration_text += "\n\tdef __init__(\n\t\tself,"
            for parameter in parameters:
                if parameter.id == "input_level":
                    graph_input.append(parameter.value)
                if not parameter.function:
                    param_set.add(parameter.id)
                if not parameter.is_stateful() and not parameter.function:
                    val = f"torch.tensor({parameter.value})"
                    ot = type(parameter.value)
                    if ot == np.ndarray:
                        val = f"torch.{torch.from_numpy(parameter.value)}"
                    declaration_text += (
                        f"\n\t\t{parameter.id} = {val}, # orig type: {ot}"
                    )

                elif parameter.is_stateful():
                    if parameter.value:
                        declaration_text += "{}=torch.tensor({}),".format(
                            parameter.id, 0
                        )
                    else:
                        if (
                            parameter.default_initial_value == "str"
                            and parameter.default_initial_value in param_set
                        ):
                            declaration_text += f"{parameter.id}=torch.tensor({parameter.default_initial_value.value}),"
                        else:
                            declaration_text += f"{parameter.id}=torch.tensor({parameter.default_initial_value}),"
            declaration_text += "\n\t):"
            declaration_text += "\n\t\tsuper().__init__()"
            for parameter in parameters:
                if not parameter.function:
                    declaration_text += f"\n\t\tself.{parameter.id}={parameter.id}"
                elif parameter.function:
                    function_set.add(parameter)

            declaration_text += "\n\t\tself.{}={}".format(
                "execution_count", "torch.tensor(0)"
            )
            # define the forward method of node
            declaration_text += "\n\tdef forward(self,"
            for input_port in input_ports:
                declaration_text += input_port.id + ","

            declaration_text += " ):"
            declaration_text += "\n\t\tself.{}=self.{}".format(
                "execution_count", "execution_count+torch.tensor(1)"
            )
            # Handles the sateful parameter in forward method
            for parameter in parameters:
                if parameter.is_stateful() and not parameter.function:
                    if parameter.value:
                        declaration_text += "\n\t\tself.{}={}".format(
                            parameter.id, sym(parameter.value)
                        )
                    elif parameter.time_derivative:
                        declaration_text += "\n\t\tself.{}=self.{}+({})".format(
                            parameter.id, parameter.id, sym(parameter.time_derivative)
                        )
                # Handles the function at forward method
                if parameter.function:
                    declaration_text += f"\n\t\t{parameter.id}="
                    if parameter.function in mdf_functions:
                        exp = mdf_functions[parameter.function]["expression_string"]
                        exp = func_args(exp, parameter.args)
                        exp = sym(exp)

                    declaration_text += exp
            # Value to be returned as output of forward method of each node
            if output_ports[0].value in not_self:
                declaration_text += "\n\t\treturn {}".format(output_ports[0].value)
            else:
                declaration_text += "\n\t\treturn {}".format(sym(output_ports[0].value))

        return declaration_text
    elif version == "mdf.0":
        for parameter in parameters:
            param_set.add(parameter.id)

        if parameters:
            declaration_text += "\n\tdef __init__(self,"
            for parameter in parameters:
                if parameter.id == "input_level":
                    graph_input.append(parameter.value)
                if not parameter.is_stateful():
                    declaration_text += (
                        f"{parameter.id}=torch.tensor({parameter.value}),"
                    )
                elif parameter.is_stateful():
                    if parameter.value:
                        declaration_text += "{}=torch.tensor({}),".format(
                            parameter.id, 0
                        )
                    else:
                        if (
                            parameter.default_initial_value == "str"
                            and parameter.default_initial_value in param_set
                        ):
                            declaration_text += f"{parameter.id}=torch.tensor({parameter.default_initial_value.value}),"
                        else:
                            declaration_text += f"{parameter.id}=torch.tensor({parameter.default_initial_value}),"
            declaration_text += "):"
            declaration_text += "\n\t\tsuper().__init__()"
            for parameter in parameters:
                declaration_text += f"\n\t\tself.{parameter.id}={parameter.id}"

            declaration_text += "\n\t\tself.{}={}".format(
                "execution_count", "torch.tensor(0)"
            )
            # define the forward method
            declaration_text += "\n\tdef forward(self,"
            for input_port in input_ports:
                declaration_text += input_port.id

            declaration_text += " ):"
            declaration_text += "\n\t\tself.{}=self.{}".format(
                "execution_count", "execution_count+torch.tensor(1)"
            )
            # Handles the sateful parameter in forward method
            for parameter in parameters:
                if parameter.is_stateful():
                    if parameter.value:
                        declaration_text += "\n\t\tself.{}={}".format(
                            parameter.id, sym(parameter.value)
                        )
                    elif parameter.time_derivative:
                        declaration_text += "\n\t\tself.{}=self.{}+({})".format(
                            parameter.id, parameter.id, sym(parameter.time_derivative)
                        )
                # Handles the function at forward method
            if functions:
                for function in functions:
                    declaration_text += f"\n\t\t{function.id}="
                    func_dic = function.function
                    for i in func_dic:
                        if i in mdf_functions:
                            exp = mdf_functions[i]["expression_string"]
                            exp = func_args(exp, func_dic[i])
                            exp = sym(exp)

                    declaration_text += exp
            # Value to be returned as output of forward method
            if output_ports[0].value in not_self:
                declaration_text += "\n\t\treturn {}".format(output_ports[0].value)
            else:
                declaration_text += "\n\t\treturn {}".format(sym(output_ports[0].value))

        return declaration_text


# Add self to parameters to make it as attribute of pytorch node
def sym(value):
    for i in param_set:
        if i in value:
            value = value.replace(i, "self." + i)
    return value


def func_args(exp, arg_dict):
    for i in arg_dict:
        if i in exp:
            exp = exp.replace(i, str(arg_dict[i]))
    return exp


# Define the forward method of the main function--graph
def generate_main_forward(
    nodes: List["node"], execution_order: List[str], d_e: Dict[str, Any]
):
    """Helper function to generate the main forward method that will specify
    the execution of the pytorch model. This requires proper ordering of module
    calls as well as preservation of variables.

    Args:
        nodes: list of nodes in the graph
        execution_order: List of nodes in the order of execution
        d_e: Weights on edges stored in dict format

    Returns: Function returns the main forward call that will be used to define pytorch
    model

    """
    node_dict = {node.id: node for node in nodes}

    d = {}
    main_forward = "\n\tdef forward(self, input):"
    if execution_order:
        for node in execution_order:
            main_forward += f"\n\t\t val_{node}=torch.zeros_like(input)"

        for node, dependency_set in d_e.items():
            if dependency_set == {}:
                main_forward += f"\n\n\t\t val_{node}=val_{node}+self.{node}()"
                d[node] = f"val_{node}"
            else:
                for k, v in dependency_set.items():
                    if v == None:
                        main_forward += (
                            f"\n\t\t val_{node}=val_{node}+self.{node}(val_{k})"
                        )
                    else:
                        main_forward += f"\n\t\t val_{node}=val_{node}+self.{node}(val_{k}*torch.tensor({v}))"
                    d[node] = f"val_{node}"

    main_forward += "\n\n\t\t return "

    for node in execution_order:
        main_forward += d[node] + ","

    return main_forward


# Create Script
def build_script(
    nodes: List["node"],
    execution_order: List[str],
    model_id1: str,
    d_e: Dict[str, Any],
    conditions,
    version: str,
):
    """Helper function to create and assemble text components necessary to specify
    module.py importable model script.  These include:

            * Module declarations
                    * Initialization of functions
                    * Definition of forward function

            * Model main call declaration:
                    * Initialization of subcomponents
                    * Forward function logic

    Args:
        nodes: list of nodes in the graph
        execution_order: List of nodes in the order of execution
        d_e: Weights on edges stored in dict format
        model_id1: id of the model to be converted
        conditions:

    Returns: complete module.py script as a formatted string

    """
    model_id = model_id1 + ".onnx"
    script = ""
    if version == "mdf.s":
        s1 = "Converted examples from MDF models version(mdf.s) to PyTorch/ONNX"
        script += f'"{s1}"'
    elif version == "mdf.0":
        s2 = "Converted examples from MDF models version(mdf.0) to PyTorch/ONNX"
        script += f'"{s2}"'
    imports_string = (
        "\nimport torch"
        "\nimport torch.nn as nn\nimport onnx\nimport onnxruntime as rt\nfrom math import *"
    )
    print(script)

    # Declarations string
    modules_declaration_text = ""
    constructor_calls = {}
    declared_module_types = set()

    for node in nodes:
        id, funcs, params, out_ports, input_ports = (
            node.id,
            node.functions,
            node.parameters,
            node.output_ports,
            node.input_ports,
        )
        node_dict = {
            "functions": funcs,
            "parameters": params,
            "output_ports": out_ports,
            "input_ports": input_ports,
        }
        declaration_text = get_module_declaration_text(
            id, node_dict, execution_order, version=version
        )
        modules_declaration_text += declaration_text

    # Build Main call
    main_call_declaration = "\n\nclass Model(nn.Module):" "\n\tdef __init__(self,"
    for node in nodes:
        main_call_declaration += f"{node.id}" + ", "
    main_call_declaration += "):" "\n\t\tsuper().__init__()"

    for idx, node in enumerate(nodes):
        main_call_declaration += f"\n\t\tself.{node.id} = {node.id}"

    # Build Main forward
    main_call_forward = generate_main_forward(nodes, execution_order, d_e)
    script += imports_string
    script += modules_declaration_text
    script += main_call_declaration
    script += main_call_forward

    if len(nodes) == 1:
        script += "f\nmodel={nodes[0].id}"
        return script
    script += "\nmodel = Model("
    for node in nodes:
        script += f"{node.id}={node.id}(),"
    script += ")"
    script += "\nmodel=torch.jit.script(model)"
    if graph_input:
        script += f"\ndummy_input =torch.tensor{tuple(graph_input)}"
    else:
        script += f"\ndummy_input =torch.tensor(0.0)"
    script += "\noutput = model(dummy_input)"
    script += f"\ntorch.onnx.export(model,dummy_input,'{model_id}',verbose=True,input_names=[],opset_version=9)"
    script += f"\nonnx_model = onnx.load('{model_id}')"
    script += "\nonnx.checker.check_model(onnx_model)"
    script += f"\nsess = rt.InferenceSession('{model_id}')"
    script += "\nres = sess.run(None, {sess.get_inputs()[0].name: dummy_input.numpy()})"
    x, y = "__main__", "Exported to PyTorch and ONNX"
    script += f"\nif __name__ == '{x}':"
    script += f"\n\tprint('{y}')"

    return script


def _generate_scripts_from_model(mdf_model: "Model", version: str) -> str:
    """Generating scripts from components of model
    Helper function to parse MDF objects from MDF json representation. Uses MDF scheduler to
    determine proper ordering of nodes, and calls `build_script`.

    Args:
        mdf_model: MDF model to be exported to PyTorch

    Returns:
        Returns dictionary of scripts where key = name of mdf model, value is string
        representation of script.

    """
    scripts = {}
    model_id1 = mdf_model.id
    for graph in mdf_model.graphs:
        nodes = graph.nodes
        evaluable_graph = EvaluableGraph(graph, verbose=False)
        enodes = evaluable_graph.enodes
        edges = evaluable_graph.ordered_edges
        try:
            conditions = evaluable_graph.conditions
        except AttributeError:
            conditions = {}

        # Use edges and nodes to construct execution order
        execution_order = []
        depend_dict = graph.dependency_dict
        d_e = {n.id: collections.defaultdict(dict) for n in graph.nodes}
        for graph in mdf_model.graphs:
            for edge in graph.edges:
                sender = graph.get_node(edge.sender)
                receiver = graph.get_node(edge.receiver)
                if edge.parameters and "weight" in edge.parameters:
                    d_e[receiver.id][sender.id] = edge.parameters["weight"]
                else:
                    d_e[receiver.id][sender.id] = None

        for idx, edge in enumerate(edges):
            if idx == 0:
                execution_order.append(edge.sender)
            execution_order.append(edge.receiver)
        if execution_order == []:
            for i in nodes:
                execution_order.append(i.id)

        # Build script
        script = build_script(
            nodes, execution_order, model_id1, d_e, conditions, version
        )
        scripts[graph.id] = script

    return scripts


# Generating PyTorch Script
def _script_to_model(script: str, model_id1: str, version: str, model_input: str):
    """
    Helper function to take the autogenerated module.py python script, and import
    it such that the pytorch model specified by this script is importable to the
    calling program.

    Returns torch.nn.Module object.
    """
    import importlib.util

    print("version", version)
    if version == "mdf.s":
        path_list = model_input.split("/")[:-2] + ["PyTorch/MDF_PyTorch/"]
        out_filename = "/".join(path_list)
        module_path = str(out_filename) + f"{model_id1}" + "_pytorch.py"
    elif version == "mdf.0":
        path_list = model_input.split("/")[:-3] + ["PyTorch/MDF_PyTorch/"]
        out_filename = "/".join(path_list)
        module_path = str(out_filename) + f"translated_{model_id1}" + "_pytorch.py"

    with open(module_path, mode="w") as f:
        f.write(script)

    torch_spec = importlib.util.spec_from_file_location("module", module_path)
    torch_module = importlib.util.module_from_spec(torch_spec)
    torch_spec.loader.exec_module(torch_module)

    model = torch_module.model

    return model


def mdf_to_pytorch(
    mdf_model: "Model", model_input: str, eval_models: bool, version: str
):
    """Function loads and returns a pytorch model for all models specified in an
    mdf file
    Args:
        mdf_model: model in MDF format
        eval_models: Set Evaluation of model to True or False
        version: MDF version
        model_input: input file name
    Returns: Returns a dictionary where key = model name, value = pytorch model object
    """
    print(
        f"Using mdf_to_pytorch to convert {mdf_model.id} from {model_input}, evaluating: {eval_models}, version {version}"
    )
    scripts = _generate_scripts_from_model(mdf_model, version)
    models = {}

    for script_name, script in scripts.items():
        model = _script_to_model(script, mdf_model.id, version, model_input)
        if eval_models:
            model.eval()
        models[script_name] = model
    return models


__all__ = ["mdf_to_pytorch"]

if __name__ == "__main__":
    from pathlib import Path
    import os

    base_path = Path(__file__).parent.parent
    filename = "examples/MDF/translation/Translated_ABCD.json"
    filename = "examples/MDF/Arrays.json"
    file_path = str((base_path / "../../.." / filename).resolve())
    model_input = file_path.replace(os.sep, "/")

    mdf_model = load_mdf(model_input)
    if "Translated" in model_input:
        pytorch_model = mdf_to_pytorch(
            mdf_model, model_input, eval_models=False, version="mdf.0"
        )
    else:
        pytorch_model = mdf_to_pytorch(
            mdf_model, model_input, eval_models=False, version="mdf.s"
        )



r"""
    The main object-oriented implementation of the MDF schema, with each core component of the `MDF specification <../Specification.html>`_
    implemented as a :code:`class`. Instances of these objects can be composed to create a representation of
    an MDF model as Python objects. These models can then be serialized and deserialized to and from JSON or YAML,
    executed via the :mod:`~modeci_mdf.execution_engine` module, or imported and exported to supported external
    environments using the :mod:`~modeci_mdf.interfaces` module.
"""
import copy
import sympy
import numpy as np

from typing import List, Tuple, Dict, Set, Any, Union, Optional

import modelspec
from modelspec import has, field, fields, optional, instance_of, in_
from modelspec.base_types import (
    Base,
    converter,
    ValueExprType,
    value_expr_types,
    value_expr_converter,
)

from ast import literal_eval as make_tuple

from modeci_mdf import MODECI_MDF_VERSION
from modeci_mdf import __version__


__all__ = [
    "Model",
    "Graph",
    "Node",
    "Function",
    "InputPort",
    "OutputPort",
    "ParameterCondition",
    "Parameter",
    "Edge",
    "ConditionSet",
    "Condition",
]


@modelspec.define(eq=False)
class MdfBase(Base):
    """
    Base class for all MDF core classes that implements common functionality.

    Attributes:
        metadata: Optional metadata field, an arbitrary dictionary of string keys and JSON serializable values.

    """

    metadata: Optional[Dict[str, Any]] = field(
        kw_only=True, default=None, validator=optional(instance_of(dict))
    )


@modelspec.define(eq=False)
class Function(MdfBase):
    r"""
    A single value which is evaluated as a function of values on :class:`InputPort`(s) and other Functions

    Attributes:
        id: The unique (for this Node) id of the function, which will be used in other :class:`~Function`s and
            the :class:`~OutputPort`s for its value
        function: Which of the in-build MDF functions (:code:`linear`, etc.). See supported functions:
            https://mdf.readthedocs.io/en/latest/api/MDF_function_specifications.html
        args: Dictionary of values for each of the arguments for the Function, e.g. if the in-built function
              is linear(slope),the args here could be {"slope":3} or {"slope":"input_port_0 + 2"}
        value: If the function is a value expression, this attribute will contain the expression and the function
            and args attributes will be None.
    """
    id: str = field(validator=instance_of(str))
    function: Optional[str] = field(
        default=None, validator=optional(instance_of((str, dict)))
    )
    args: Optional[Dict[str, Any]] = field(default=None)
    value: Optional[ValueExprType] = field(
        default=None, validator=optional(instance_of(value_expr_types))
    )

    @classmethod
    def _parse_cattr_structure(cls, o):
        if "function" in o:
            if isinstance(o["function"], dict):
                # TODO: handle args already present and conflicting?
                func_name = list(o["function"].keys())[0]
                o["args"] = o["function"][func_name]
                o["function"] = func_name
        return o


@modelspec.define(eq=False)
class InputPort(MdfBase):
    r"""
    The :class:`InputPort` is an attribute of a Node which allows external information to be input to the Node

    Attributes:
        id: The unique (for this Node) id of the input port,
        shape: The shape of the input port. This uses the same syntax as numpy ndarray shapes
            (e.g., :code:`numpy.zeros(shape)` would produce an array with the correct shape
        type: The data type of the input received at a port.

    """
    id: str = field(validator=instance_of(str))
    shape: Optional[Tuple[int, ...]] = field(
        validator=optional(instance_of(tuple)),
        default=None,
        converter=lambda x: make_tuple(x) if type(x) == str else x,
    )
    type: Optional[str] = field(validator=optional(instance_of(str)), default=None)


@modelspec.define(eq=False)
class OutputPort(MdfBase):
    r"""
    The :class:`OutputPort` is an attribute of a :class:`Node` which exports information to another :class:`Node`
    connected by an :class:`Edge`

    Attributes:
        id: Unique identifier for the output port.
        value: The value of the :class:`OutputPort` in terms of the :class:`InputPort`, :class:`Function` values, and
            :class:`Parameter` values.
        shape: The shape of the output port. This uses the same syntax as numpy ndarray shapes
            (e.g., :code:`numpy.zeros(shape)` would produce an array with the correct shape
        type: The data type of the output sent by a port.
    """
    id: str = field(validator=instance_of(str))
    value: Optional[str] = field(validator=optional(instance_of(str)), default=None)
    shape: Optional[Tuple[int, ...]] = field(
        validator=optional(instance_of(tuple)),
        default=None,
        converter=lambda x: make_tuple(x) if type(x) == str else x,
    )
    type: Optional[str] = field(validator=optional(instance_of(str)), default=None)


@modelspec.define(eq=False)
class ParameterCondition(Base):
    r"""
    A condition to test on a Node's parameters, which if true, sets the value of this Parameter

    Attributes:
        id: A unique identifier for the ParameterCondition
        test: The boolean expression to evaluate
        value: The new value of the Parameter if the test is true
    """
    id: str = field(validator=instance_of(str))
    test: Optional[ValueExprType] = field(default=None)
    value: Optional[ValueExprType] = field(default=None)


@modelspec.define(eq=False)
class Parameter(MdfBase):
    r"""
    A parameter of the :class:`Node`, which can be: 1) a specific fixed :code:`value` (a constant (int/float) or an array) 2) a string expression for the :code:`value`
    referencing other named :class:`Parameter`(s). which may be stateful (i.e. can change value over multiple executions of the :class:`Node`); 3) be evaluated by an
    inbuilt :code:`function` with :code:`args`; 4) or change from a :code:`default_initial_value` with a :code:`time_derivative`.

    Attributes:
        value: The next value of the parameter, in terms of the inputs, functions and PREVIOUS parameter values
        default_initial_value: The initial value of the parameter, only used when parameter is stateful.
        time_derivative: How the parameter changes with time, i.e. ds/dt. Units of time are seconds.
        function: Which of the in-build MDF functions (linear etc.) this uses, See
        https://mdf.readthedocs.io/en/latest/api/MDF_function_specifications.html
        args: Dictionary of values for each of the arguments for the function of the parameter,
            e.g. if the in-build function is :code:`linear(slope)`, the args here could be :code:`{"slope": 3}` or
            :code:`{"slope": "input_port_0 + 2"}`
        conditions: Parameter specific conditions
    """

    id: str = field(validator=instance_of(str))
    value: Optional[ValueExprType] = field(default=None)
    default_initial_value: Optional[ValueExprType] = field(default=None)
    time_derivative: Optional[str] = field(
        validator=optional(instance_of(str)), default=None
    )
    function: Optional[str] = field(validator=optional(instance_of(str)), default=None)
    args: Optional[Dict[str, Any]] = field(
        validator=optional(instance_of(dict)), default=None
    )
    conditions: List[ParameterCondition] = field(
        factory=list, validator=instance_of(list)
    )

    def summary(self):
        """
        Short summary of Parameter...
        """
        info = "Parameter: {} = {} (stateful: {})".format(
            self.id,
            self.value,
            self.is_stateful(),
        )

        return info

    def is_stateful(self) -> bool:
        """
        Is the parameter stateful?

        A parameter is considered stateful if it has a :code:`time_derivative`, :code:`defualt_initial_value`, or it's
        id is referenced in its value expression.

        Returns:
            :code:`True` if stateful, `False` if not.
        """
        from modeci_mdf.execution_engine import parse_str_as_list

        if self.time_derivative is not None:
            return True
        if self.default_initial_value is not None:
            return True
        if self.value is not None and type(self.value) == str:
            # If we are dealing with a list of symbols, each must treated separately
            if self.value[0] == "[" and self.value[-1] == "]":
                # Use the Python interpreter to parse this into a List[str]
                arg_expr_list = parse_str_as_list(self.value)
            else:
                arg_expr_list = [self.value]

            req_vars = []

            for e in arg_expr_list:
                param_expr = sympy.simplify(e)
                req_vars.extend([str(s) for s in param_expr.free_symbols])
            sf = self.id in req_vars
            """
            print(
                "Checking whether %s is stateful, %s: %s"
                % (self, param_expr.free_symbols, sf)
            )"""
            return sf

        return False

    @staticmethod
    def _unstructure(o: "Parameter"):
        """A custom unstructuring hook for parameters that removes default values of None from the output."""

        # Unstructure the parameter and any value expressions it contains.
        d = value_expr_converter.unstructure(o)

        # Remove any fields that are None by default
        d = {
            name: val
            for name, val in d.items()
            if not ((val is None) or (type(val) == list and len(val) == 0))
        }

        return d


@modelspec.define(eq=False)
class Node(MdfBase):
    r"""
    A self contained unit of evaluation receiving input from other nodes on :class:`InputPort`(s).
    The values from these are processed via a number of :class:`Function`(s) and one or more final values
    are calculated on the :class:`OutputPort`(s)

    Attributes:
        id: A unique identifier for the node.
        input_ports: Dictionary of the :class:`InputPort` objects in the Node
        parameters: Dictionary of :class:`Parameter`(s) for the node
        functions: The :class:`Function`(s) for computation the node
        output_ports: The :class:`OutputPort`(s) containing evaluated quantities from the node
    """

    id: str = field(validator=instance_of(str))
    input_ports: List[InputPort] = field(factory=list, validator=instance_of(list))
    functions: List[Function] = field(factory=list, validator=instance_of(list))
    parameters: List[Parameter] = field(factory=list, validator=instance_of(list))
    output_ports: List[OutputPort] = field(factory=list, validator=instance_of(list))

    def get_parameter(self, id: str) -> Union[Parameter, None]:
        r"""Get a parameter by its string :code:`id`

        Args:
            id: The unique string id of the :class:`Parameter`

        Returns:
            The :class:`Parameter` object stored on this node. :code:`None` if not found.
        """
        for p in self.parameters:
            if p.id == id:
                return p

        return None

    def get_input_port(self, id: str) -> "InputPort":
        """Retrieve :class:`InputPort` object corresponding to the given id
        Args:
            id: Unique identifier of class:`InputPort` object
        Returns:
            class:`InputPort` object if the entered id matches with the id of class:`InputPort` present in the class:`Node`
        """
        for ip in self.input_ports:
            if id == ip.id:
                return ip

    def get_output_port(self, id: str) -> "OutputPort":
        """Retrieve :class:`OutputPort` object corresponding to the given id
        Args:
            id: Unique identifier of class:`OutputPort` object
        Returns:
            class:`OutputPort` object if the entered id matches with the id of class:`OutputPort` present in the class:`Node`
        """
        for op in self.output_ports:
            if id == op.id:
                return op


@modelspec.define(eq=False)
class Edge(MdfBase):
    r"""
    An :class:`Edge` is an attribute of a :class:`Graph` that transmits computational results from a sender's
    :class:`OutputPort` to a receiver's :class:`InputPort`.

    Attributes:
        id: A unique string identifier for this edge.
        sender: The :code:`id` of the :class:`~Node` which is the source of the edge.
        receiver: The :code:`id` of the :class:`~Node` which is the target of the edge.
        sender_port: The id of the :class:`~OutputPort` on the sender :class:`~Node`, whose value should be sent to the
            :code:`receiver_port`
        receiver_port: The id of the InputPort on the receiver :class:`~Node`
        parameters: Dictionary of parameters for the edge.
    """
    id: str = field(validator=instance_of(str))
    sender: str = field(validator=instance_of(str))
    receiver: str = field(validator=instance_of(str))
    sender_port: str = field(validator=instance_of(str))
    receiver_port: str = field(validator=instance_of(str))
    parameters: Optional[Dict[str, Any]] = field(
        validator=optional(instance_of(dict)), default=None
    )


@modelspec.define(eq=False)
class Condition(MdfBase):
    r"""A set of descriptors which specifies conditional execution of Nodes to meet complex execution requirements.

    Attributes:
        type: The type of :class:`Condition` from the library
        kwargs: The dictionary of keyword arguments needed to evaluate the :class:`Condition`

    """
    type: str = field(validator=instance_of(str))
    kwargs: Optional[Dict[str, Any]] = field(
        validator=optional(instance_of(dict)), default=None
    )

    def __init__(
        self,
        type: Optional[str] = None,
        **kwargs: Optional[Dict[str, Any]],
    ):
        # We need to right our own __init__ in this case because the current API for Condition requires saving
        # kwargs. This code is very attrs specific and hacky. Wish there was a better way to do this.

        # Remove any field from kwargs that is pre-defined field on the MDF Condition class.
        for a in self.__attrs_attrs__:
            if a.name != "kwargs":
                kwargs.pop(a.name, None)

        # If there is a kwargs argument inside the kwargs argument (this happens when loading MDF from JSON because)
        # kwargs is stored in JSON as a simple dict in a field called kwargs.
        kwargs.update(kwargs.pop("kwargs", {}))

        self.__attrs_init__(type, kwargs)

    @classmethod
    def _parse_cattr_structure(cls, o):
        def _parse_as_condition(arg):
            try:
                return Condition(**cls._parse_cattr_structure(arg))
            except (TypeError, ValueError):
                return arg

        if "kwargs" in o:
            for k in o["kwargs"]:
                if isinstance(o["kwargs"][k], list):
                    o["kwargs"][k] = [_parse_as_condition(a) for a in o["kwargs"][k]]
                else:
                    try:
                        # read Model object as just object id
                        o["kwargs"][k] = o["kwargs"][k]["id"]
                    except (KeyError, TypeError):
                        o["kwargs"][k] = _parse_as_condition(o["kwargs"][k])

        return o

    @classmethod
    def _parse_cattr_unstructure(cls, o):
        def _clean_nested_Condition_objects(cond):
            res = copy.copy(cond)

            for k, v in res.kwargs.items():
                if isinstance(v, Condition):
                    res.kwargs[k] = _clean_nested_Condition_objects(v)
                elif isinstance(v, Base):
                    # replace mdf objects with their id to avoid making duplicates
                    res.kwargs[k] = v.id

            return res

        return _clean_nested_Condition_objects(o)


@modelspec.define(eq=False)
class ConditionSet(MdfBase):
    r"""
    Specifies the non-default pattern of execution of Nodes

    Attributes:
        node_specific: A dictionary mapping nodes to any non-default run conditions
        termination: A dictionary mapping time scales of model execution to conditions indicating when they end
    """
    node_specific: Optional[Dict[str, Condition]] = field(
        validator=optional(instance_of(dict)), default=None
    )
    termination: Optional[Dict[str, Condition]] = field(
        validator=optional(instance_of(dict)), default=None
    )

    @classmethod
    def _parse_cattr_unstructure(cls, o):
        res = copy.copy(o)
        if res.termination is not None:
            res.termination = {str(k): v for k, v in res.termination.items()}

        return res


@modelspec.define(eq=False)
class Graph(MdfBase):
    r"""
    A directed graph consisting of Node(s) connected via Edge(s)

    Attributes:
        id: A unique identifier for this Graph
        nodes: One or more :class:`Node`(s) present in the graph
        edges: Zero or more :class:`Edge`(s) present in the graph
        parameters: Dictionary of global parameters for the Graph
        conditions: The ConditionSet stored as dictionary for scheduling of the Graph
    """
    id: str = field(validator=instance_of(str))
    nodes: List[Node] = field(factory=list)
    edges: List[Edge] = field(factory=list)
    parameters: Optional[Dict[str, Any]] = None
    conditions: Optional[ConditionSet] = None

    def get_node(self, id: str) -> Union[Node, None]:
        """Retrieve Node object corresponding to the given id

        Args:
            id: Unique identifier of Node object

        Returns:
            :class:`Node` object if the entered :code:`id` matches with the :code:`id` of node present in the
            :class:`~Graph`. :code:`None` if a node is not found with that id .
        """
        for node in self.nodes:
            if id == node.id:
                return node

        return None

    @property
    def dependency_dict(self) -> Dict[Node, Set[Node]]:
        """Returns the dependency among nodes as dictionary

        Key: receiver, Value: Set of senders imparting information to the receiver

        Returns:
            Returns the dependency dictionary
        """
        # assumes no cycles, need to develop a way to prune if cyclic
        # graphs are to be supported
        dependencies = {n: set() for n in self.nodes}

        for edge in self.edges:
            sender = self.get_node(edge.sender)
            receiver = self.get_node(edge.receiver)

            dependencies[receiver].add(sender)

        return dependencies

    @property
    def inputs(self: "Graph") -> List[Tuple[Node, InputPort]]:
        """
        Enumerate all Node-InputPort pairs that specify no incoming edge.
        These are input ports for the graph itself and must be provided values to evaluate

        Returns:
            A list of Node, InputPort tuples
        """

        # Get all input ports
        all_ips = [(node.id, ip.id) for node in self.nodes for ip in node.input_ports]

        # Get all receiver ports
        all_receiver_ports = {(e.receiver, e.receiver_port) for e in self.edges}

        # Find any input ports that aren't receiving values from an edge
        return list(filter(lambda x: x not in all_receiver_ports, all_ips))


@modelspec.define(eq=False)
class Model(MdfBase):
    r"""
    The top level construct in MDF is Model, which may contain multiple :class:`Graph` objects and model attribute(s)

    Attributes:
        id: A unique identifier for this Model
        graphs: The collection of graphs that make up the MDF model.
        format: Information on the version of MDF used in this file
        generating_application: Information on what application generated/saved this file
        onnx_opset_version: The ONNX opset used for any ONNX functions in this model.

    """
    id: str = field(validator=instance_of(str))
    graphs: List[Graph] = field(factory=list)
    format: str = field(
        default=f"ModECI MDF v{MODECI_MDF_VERSION}", metadata={"omit_if_default": False}
    )
    generating_application: str = field(
        default=f"Python modeci-mdf v{__version__}", metadata={"omit_if_default": False}
    )
    onnx_opset_version: Optional[str] = field(default=None)

    def to_graph_image(
        self,
        engine: str = "dot",
        output_format: str = "png",
        view_on_render: bool = False,
        level: int = 2,
        filename_root: Optional[str] = None,
        only_warn_on_fail: bool = False,
    ):
        """Convert MDF graph to an image (png or svg) using the Graphviz export

        Args:
            engine: dot or other Graphviz formats
            output_format: e.g. png (default) or svg
            view_on_render: if True, will open generated image in system viewer
            level: 1,2,3, depending on how much detail to include
            filename_root: will change name of file generated to filename_root.png, etc.
            only_warn_on_fail: just give a warning if this fails, e.g. no dot executable. Useful for preventing errors in automated tests
        """
        from modeci_mdf.interfaces.graphviz.exporter import mdf_to_graphviz

        try:
            mdf_to_graphviz(
                self.graphs[0],
                engine=engine,
                output_format=output_format,
                view_on_render=view_on_render,
                level=level,
                filename_root=filename_root,
            )

        except Exception as e:
            if only_warn_on_fail:
                print(
                    "Failure to generate image! Ensure Graphviz executables (dot etc.) are installed on native system. Error: \n%s"
                    % e
                )
            else:
                raise (e)


# Add a special hook for handling unstructuring of Parameters and Functions
converter.register_unstructure_hook(Parameter, Parameter._unstructure)


def parsed_structure_factory(cl):
    base_structure = converter._structure_func.dispatch(cl)

    def new_structure(obj, t):
        obj = cl._parse_cattr_structure(obj)
        return base_structure(obj, t)

    return new_structure


def parsed_unstructure_factory(cl):
    base_unstructure = converter._unstructure_func.dispatch(cl)

    def new_unstructure(obj):
        obj = cl._parse_cattr_unstructure(obj)
        return base_unstructure(obj)

    return new_unstructure


for k, v in list(locals().items()):
    # add structure hook for MdfBase classes where present
    try:
        v._parse_cattr_structure
    except AttributeError:
        pass
    else:
        converter.register_structure_hook(v, parsed_structure_factory(v))

    # add unstructure hook for MdfBase classes where present
    try:
        v._parse_cattr_unstructure
    except AttributeError:
        pass
    else:
        converter.register_unstructure_hook(v, parsed_unstructure_factory(v))



"""
Simple export of MDF to GraphViz for generating graphics.

Work in progress...
"""

import sys

import graphviz
import numpy as np

from modeci_mdf.functions.standard import mdf_functions

from modeci_mdf.utils import color_rgb_to_hex

from modelspec.utils import _val_info

engines = {
    "d": "dot",
    "c": "circo",
    "n": "neato",
    "t": "twopi",
    "f": "fdp",
    "s": "sfdp",
    "p": "patchwork",
}

NO_VIEW = "-noview"

LEVEL_1 = 1
LEVEL_2 = 2
LEVEL_3 = 3

COLOR_MAIN = "#444444"
# COLOR_BG_MAIN = "#999911"
COLOR_LABEL = "#666666"
COLOR_NUM = "#444444"
COLOR_PARAM = "#1666ff"
COLOR_INPUT = "#188855"
COLOR_FUNC = "#441199"
COLOR_OUTPUT = "#cc3355"
COLOR_COND = "#ffa1d"


def format_label(s):
    # return f'<font color="{COLOR_LABEL}"><b>{s}</b></font></td><td>'
    return ""


def format_num(s):
    if type(s) == np.ndarray:
        ss = "%s" % (np.array2string(s, threshold=4, edgeitems=1))
        info = f" (NP {s.shape} {s.dtype})"
    else:
        ss = s
        info = ""
    return f'<font color="{COLOR_NUM}"><b>{ss}</b>{info}</font>'


def format_bold(s, use_bold=True):
    return f"<b>{s}</b>" if use_bold else s


def format_param(s):
    return f'<font color="{COLOR_PARAM}">{s}</font>'


def format_input(s):
    return f'<font color="{COLOR_INPUT}">{s}</font>'


def format_function(s):
    return f'<font color="{COLOR_FUNC}">{s}</font>'


def format_standard_func(s):
    return "<i>%s</i>" % (s)


def format_standard_func_long(s):
    return "<i>%s</i>" % (s)


def format_output(s):
    return f'<font color="{COLOR_OUTPUT}">{s}</font>'


def format_condition(s):
    return f'<font color="{COLOR_COND}">{s}</font>'


def match_in_expr(expr, node):

    if type(expr) != str:
        return "%s" % _val_info(expr)
    else:
        # print("Checking %s" % (expr))

        expr = " %s " % expr

        def _replace_var(v, expr, format_method):
            # print(f"Replacing {v} in {expr}")
            if expr == v:
                return format_method(expr)
            can_start = [" ", "+", "-", "*", "/", "("]
            can_end = [" ", "+", "-", "*", "/", ")"]
            for s in can_start:
                for e in can_end:
                    expr = expr.replace(s + v + e, s + format_method(v) + e)
            return expr

        for p in node.parameters:
            expr = _replace_var(p.id, expr, format_param)

        for ip in node.input_ports:
            expr = _replace_var(ip.id, expr, format_input)

        for f in node.functions:
            expr = _replace_var(f.id, expr, format_function)

        for op in node.output_ports:
            expr = _replace_var(op.id, expr, format_output)

        # print("Checked %s" % (expr))

        return expr.strip()


def mdf_to_graphviz(
    mdf_graph,
    engine="dot",
    output_format="png",
    view_on_render=False,
    level=LEVEL_2,
    filename_root=None,
):

    DEFAULT_POP_SHAPE = "ellipse"
    DEFAULT_ARROW_SHAPE = "empty"

    print(
        f"Converting MDF graph: {mdf_graph.id} to graphviz (level: {level}, format: {output_format})"
    )

    graph = graphviz.Digraph(
        mdf_graph.id,
        filename="%s.gv" % mdf_graph.id if not filename_root else filename_root,
        engine=engine,
        format=output_format,
    )

    for node in mdf_graph.nodes:
        print("    Node: %s" % node.id)
        color = COLOR_MAIN
        penwidth = "1"
        # bg_color = COLOR_BG_MAIN

        if node.metadata is not None:
            if "color" in node.metadata:
                color = color_rgb_to_hex(node.metadata["color"])
                penwidth = "2"

        graph.attr(
            "node",
            color=color,
            style="rounded",
            shape="box",
            fontcolor=COLOR_MAIN,
            penwidth=penwidth,
        )
        info = '<table border="0" cellborder="0">'
        info += '<tr><td colspan="2"><b>%s</b></td></tr>' % (node.id)

        if node.metadata is not None and level >= LEVEL_3:

            info += "<tr><td>%s" % format_label("METADATA")

            for m in node.metadata:
                info += format_standard_func_long("{} = {}".format(m, node.metadata[m]))

            info += "</td></tr>"

        if level >= LEVEL_2:

            if node.input_ports and len(node.input_ports) > 0:
                for ip in node.input_ports:
                    info += "<tr><td>{}{} {}</td></tr>".format(
                        format_label("IN"),
                        format_input(ip.id),
                        "(shape: %s)" % str(ip.shape)
                        if level >= LEVEL_2 and ip.shape is not None
                        else "",
                    )

            if node.parameters and len(node.parameters) > 0:

                for p in node.parameters:
                    try:
                        stateful = p.is_stateful()
                    except:
                        stateful = False
                    if p.function is not None:
                        argstr = (
                            ", ".join(
                                [match_in_expr(str(p.args[a]), node) for a in p.args]
                            )
                            if p.args
                            else "???"
                        )
                        info += "<tr><td>{}{} = {}({})</td></tr>".format(
                            format_label(" "),
                            format_bold(format_param(p.id), stateful),
                            format_standard_func(p.function),
                            argstr,
                        )
                        if level >= LEVEL_3:
                            func_info = mdf_functions[p.function]
                            info += "<tr><td>%s</td></tr>" % (
                                format_standard_func_long(
                                    "%s(%s) = %s"
                                    % (
                                        p.function,
                                        ", ".join([a for a in p.args]),
                                        func_info["expression_string"],
                                    )
                                )
                            )
                    else:
                        v = ""
                        if p.value is not None:
                            val = match_in_expr(p.value, node)
                            v += val
                        if p.default_initial_value is not None:
                            v += "<i>def init value:</i> %s" % match_in_expr(
                                p.default_initial_value, node
                            )
                        if p.time_derivative is not None:
                            v += ", <i>d/dt:</i> %s" % match_in_expr(
                                p.time_derivative, node
                            )
                        for cond in p.conditions:
                            test = cond.test.replace(">", "&gt;").replace("<", "&lt;")
                            v += "<br/><i>{}: </i>IF {} THEN {}={}".format(
                                cond.id,
                                match_in_expr(test, node),
                                format_param(p.id),
                                match_in_expr(cond.value, node),
                            )
                        info += "<tr><td>{}{} = {}</td></tr>".format(
                            format_label(" "),
                            format_bold(format_param(p.id), stateful),
                            v,
                        )

            if node.functions and len(node.functions) > 0:
                for f in node.functions:
                    if f.function is not None:
                        argstr = (
                            ", ".join(
                                [match_in_expr(str(f.args[a]), node) for a in f.args]
                            )
                            if f.args
                            else "???"
                        )
                        info += "<tr><td>{}{} = {}({})</td></tr>".format(
                            format_label("FUNC"),
                            format_function(f.id),
                            format_standard_func(f.function),
                            argstr,
                        )
                        if level >= LEVEL_3:
                            func_info = mdf_functions[f.function]
                            info += '<tr><td colspan="2">%s</td></tr>' % (
                                format_standard_func_long(
                                    "%s(%s) = %s"
                                    % (
                                        f.function,
                                        ", ".join([a for a in f.args]),
                                        func_info["expression_string"],
                                    )
                                )
                            )
                    elif f.value is not None:
                        argstr = "("
                        if f.args:
                            for a in f.args:
                                argstr += "{}={}, ".format(a, f.args[a])
                        else:
                            argstr += " - no args -  "
                        argstr = argstr[:-2] + ")"

                        info += "<tr><td>{}{} = {} {}</td></tr>".format(
                            format_label("FUNC"),
                            format_function(f.id),
                            format_standard_func(f.value),
                            argstr,
                        )
            if mdf_graph.conditions and mdf_graph.conditions.node_specific:
                ns = mdf_graph.conditions.node_specific[node.id]
                info += "<tr><td>{}{}={} ".format(
                    format_label(" "),
                    format_condition("condition"),
                    ns.type,
                )
                args = ns.kwargs
                if args:
                    for con in args:
                        nn = format_num(args[con])
                        breaker = "<br/>"
                        info += "{} = {}{}".format(
                            format_condition(con),
                            nn,
                            breaker if len(info.split(breaker)[-1]) > 500 else ";    ",
                        )
                    info = info[:-5]

                info += "</td></tr>"

            if node.output_ports and len(node.output_ports) > 0:
                for op in node.output_ports:
                    info += "<tr><td>{}{} = {} {}</td></tr>".format(
                        format_label("OUT"),
                        format_output(op.id),
                        match_in_expr(op.value, node),
                        "(shape: %s)" % op.shape
                        if op.shape is not None
                        else ""
                        if level >= LEVEL_2 and op.shape is not None
                        else "",
                    )

        info += "</table>"

        graph.node(node.id, label="<%s>" % info)

    for edge in mdf_graph.edges:
        print(f"    Edge: {edge.id} connects {edge.sender} to {edge.receiver}")

        label = "%s" % edge.id
        if level >= LEVEL_2:
            label += " ({} -&gt; {})".format(
                format_output(edge.sender_port),
                format_input(edge.receiver_port),
            )
            if edge.parameters:
                for p in edge.parameters:
                    label += "<br/>{}: <b>{}</b>".format(
                        p, format_num(edge.parameters[p])
                    )

        graph.edge(
            edge.sender,
            edge.receiver,
            arrowhead=DEFAULT_ARROW_SHAPE,
            label="<%s>" % label if level >= LEVEL_2 else "",
        )

    if view_on_render:
        graph.view()
    else:
        name = graph.render()
        print("Written graph image to: %s" % name)


if __name__ == "__main__":

    from modeci_mdf.utils import load_mdf, print_summary

    verbose = True

    if len(sys.argv) < 3:
        print(
            "Usage:\n\n  python graphviz.py MDF_JSON_FILE level [%s]\n\n" % NO_VIEW
            + "where level = 1, 2 or 3. Include %s to supress viewing generated graph on render\n"
            % NO_VIEW
        )
        exit()

    example = sys.argv[1]
    view = NO_VIEW not in sys.argv

    model = load_mdf(example)

    mod_graph = model.graphs[0]

    print("------------------")
    print("Loaded Graph:")
    print_summary(mod_graph)

    print("------------------")
    # nmllite_file = example.replace('.json','.nmllite.json')
    mdf_to_graphviz(
        mod_graph, engine=engines["d"], view_on_render=view, level=int(sys.argv[2])
    )
